"""Tests for WebRTC utility."""

from google_nest_sdm.webrtc_util import (
    SDPDirection,
    SDPMediaKind,
    _add_foundation_to_candidates,
    _get_media_direction,
    _update_direction_in_answer,
    fix_mozilla_sdp_answer,
)


def test_fix_mozilla_sdp_answer() -> None:
    """Test the fix in the SDP for Firefox."""
    firefox_offer_sdp = (
        "v=0\r\n"
        "o=mozilla...THIS_IS_SDPARTA-99.0 137092584186714854 0 IN IP4 0.0.0.0\r\n"
        "m=audio 9 UDP/TLS/RTP/SAVPF 109 9 0 8 101\r\n"
        "c=IN IP4 0.0.0.0\r\n"
        "a=recvonly\r\n"
        "m=video 9 UDP/TLS/RTP/SAVPF 120 124 121 125 126 127 97 98 123 122 119\r\n"
        "c=IN IP4 0.0.0.0\r\n"
        "a=recvonly\r\n"
        "m=application 9 UDP/DTLS/SCTP webrtc-datachannel\r\n"
        "c=IN IP4 0.0.0.0\r\n"
        "a=sendrecv\r\n"
    )
    answer_sdp = (
        "v=0\r\n"
        "o=- 0 2 IN IP4 127.0.0.1\r\n"
        "m=audio 19305 UDP/TLS/RTP/SAVPF 109\r\n"
        "c=IN IP4 74.125.247.118\r\n"
        "a=rtcp:9 IN IP4 0.0.0.0\r\n"
        "a=candidate: 1 udp 2113939711 2001:4860:4864:4::118 19305 typ host generation 0\r\n"
        "a=candidate: 1 tcp 2113939710 2001:4860:4864:4::118 19305 typ host tcptype passive generation 0\r\n"
        "a=candidate: 1 ssltcp 2113939709 2001:4860:4864:4::118 443 typ host generation 0\r\n"
        "a=candidate: 1 udp 2113932031 74.125.247.118 19305 typ host generation 0\r\n"
        "a=candidate: 1 tcp 2113932030 74.125.247.118 19305 typ host tcptype passive generation 0\r\n"
        "a=candidate: 1 ssltcp 2113932029 74.125.247.118 443 typ host generation 0\r\n"
        "a=sendrecv\r\n"
        "m=video 9 UDP/TLS/RTP/SAVPF 126 127\r\n"
        "c=IN IP4 0.0.0.0\r\n"
        "a=rtcp:9 IN IP4 0.0.0.0\r\n"
        "a=sendrecv\r\n"
        "m=application 9 DTLS/SCTP 5000\r\n"
        "c=IN IP4 0.0.0.0\r\n"
    )
    expected_answer_sdp = (
        "v=0\r\n"
        "o=- 0 2 IN IP4 127.0.0.1\r\n"
        "m=audio 19305 UDP/TLS/RTP/SAVPF 109\r\n"
        "c=IN IP4 74.125.247.118\r\n"
        "a=rtcp:9 IN IP4 0.0.0.0\r\n"
        "a=candidate:1 1 udp 2113939711 2001:4860:4864:4::118 19305 typ host generation 0\r\n"
        "a=candidate:2 1 tcp 2113939710 2001:4860:4864:4::118 19305 typ host tcptype passive generation 0\r\n"
        "a=candidate:3 1 ssltcp 2113939709 2001:4860:4864:4::118 443 typ host generation 0\r\n"
        "a=candidate:4 1 udp 2113932031 74.125.247.118 19305 typ host generation 0\r\n"
        "a=candidate:5 1 tcp 2113932030 74.125.247.118 19305 typ host tcptype passive generation 0\r\n"
        "a=candidate:6 1 ssltcp 2113932029 74.125.247.118 443 typ host generation 0\r\n"
        "a=sendonly\r\n"
        "m=video 9 UDP/TLS/RTP/SAVPF 126 127\r\n"
        "c=IN IP4 0.0.0.0\r\n"
        "a=rtcp:9 IN IP4 0.0.0.0\r\n"
        "a=sendonly\r\n"
        "m=application 9 DTLS/SCTP 5000\r\n"
        "c=IN IP4 0.0.0.0\r\n"
    )
    chrome_offer_sdp = (
        "v=0\r\n"
        "o=- 6714414228100263102 2 IN IP4 127.0.0.1\r\n"
        "m=audio 9 UDP/TLS/RTP/SAVPF 111 63 9 0 8 13 110 126\r\n"
        "c=IN IP4 0.0.0.0\r\n"
        "a=recvonly\r\n"
        "m=video 9 UDP/TLS/RTP/SAVPF 96 97 98 99 100 101 35 36 37 38 102 103 104 105 106 107 108 109 127 125 39 40 41 42 43 44 45 46 47 48 112 113 114 115 116 117 118 49\r\n"
        "c=IN IP4 0.0.0.0\r\n"
        "a=recvonly\r\n"
        "m=application 9 UDP/DTLS/SCTP webrtc-datachannel\r\n"
        "c=IN IP4 0.0.0.0\r\n"
    )
    fixed_sdp = fix_mozilla_sdp_answer(firefox_offer_sdp, answer_sdp)
    assert fixed_sdp == expected_answer_sdp

    fixed_sdp = fix_mozilla_sdp_answer(chrome_offer_sdp, answer_sdp)
    assert fixed_sdp == answer_sdp


def test_get_media_direction() -> None:
    """Test getting the direction in the SDP."""
    sdp = (
        "v=0\r\n"
        "o=- 123456 654321 IN IP4 127.0.0.1\r\n"
        "s=Test\r\n"
        "c=IN IP4 127.0.0.1\r\n"
        "t=0 0\r\n"
        "m=audio 49170 RTP/AVP 0\r\n"
        "a=rtpmap:0 PCMU/8000\r\n"
        "a=sendrecv\r\n"
        "m=video 51372 RTP/AVP 96\r\n"
        "a=rtpmap:96 H264/90000\r\n"
        "a=sendonly\r\n"
    )

    direction = _get_media_direction(sdp, SDPMediaKind.AUDIO)
    assert direction == SDPDirection.SENDRECV
    direction = _get_media_direction(sdp, SDPMediaKind.VIDEO)
    assert direction == SDPDirection.SENDONLY
    direction = _get_media_direction(sdp, SDPMediaKind.APPLICATION)
    assert direction is None


def test_update_direction_in_answer() -> None:
    """Test updating the direction in the SDP answer."""
    original_sdp = (
        "v=0\r\n"
        "o=- 123456 654321 IN IP4 127.0.0.1\r\n"
        "s=Test\r\n"
        "c=IN IP4 127.0.0.1\r\n"
        "t=0 0\r\n"
        "m=audio 49170 RTP/AVP 0\r\n"
        "a=rtpmap:0 PCMU/8000\r\n"
        "a=sendrecv\r\n"  # Existing direction
        "m=video 51372 RTP/AVP 96\r\n"
        "a=rtpmap:96 H264/90000\r\n"
        "a=sendrecv\r\n"
    )

    # Expected result after changing the audio direction
    expected_sdp = (
        "v=0\r\n"
        "o=- 123456 654321 IN IP4 127.0.0.1\r\n"
        "s=Test\r\n"
        "c=IN IP4 127.0.0.1\r\n"
        "t=0 0\r\n"
        "m=audio 49170 RTP/AVP 0\r\n"
        "a=rtpmap:0 PCMU/8000\r\n"
        "a=sendonly\r\n"  # Updated direction
        "m=video 51372 RTP/AVP 96\r\n"
        "a=rtpmap:96 H264/90000\r\n"
        "a=sendrecv\r\n"
    )

    new_sdp = _update_direction_in_answer(
        original_sdp, SDPMediaKind.AUDIO, SDPDirection.SENDRECV, SDPDirection.SENDONLY
    )

    assert new_sdp == expected_sdp


def test_add_foundation_to_candidates() -> None:
    """Test adding a foundation value to ICE candidates."""
    original_sdp = (
        "v=0\r\n"
        "o=- 123456 654321 IN IP4 127.0.0.1\r\n"
        "s=Test\r\n"
        "c=IN IP4 127.0.0.1\r\n"
        "t=0 0\r\n"
        "a=candidate: 1 UDP 2122260223 192.168.0.1 49170 typ host\r\n"
        "a=candidate: 1 UDP 2122260223 192.168.0.1 51372 typ host\r\n"
    )

    expected_sdp = (
        "v=0\r\n"
        "o=- 123456 654321 IN IP4 127.0.0.1\r\n"
        "s=Test\r\n"
        "c=IN IP4 127.0.0.1\r\n"
        "t=0 0\r\n"
        "a=candidate:1 1 UDP 2122260223 192.168.0.1 49170 typ host\r\n"
        "a=candidate:2 1 UDP 2122260223 192.168.0.1 51372 typ host\r\n"
    )

    new_sdp = _add_foundation_to_candidates(original_sdp)

    assert new_sdp == expected_sdp



"""Base model for all nest trait based classes."""

from __future__ import annotations

from dataclasses import dataclass, fields
from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig
from typing import Any, Mapping, Self


TRAITS = "traits"
SDM_PREFIX = "sdm."


@dataclass
class TraitDataClass(DataClassDictMixin):
    """Base model for API objects that are trait based.

    This is meant to be subclasses by the model definitions.
    """

    @classmethod
    def parse_trait_object(cls, raw_data: Mapping[str, Any]) -> Self:
        """Parse a new dataclass"""
        return cls.from_dict(
            {
                **raw_data,
                **raw_data.get(TRAITS, {}),
            }
        )

    @property
    def traits(self) -> dict[str, Any]:
        """Return a trait mixin on None."""
        return {
            alias: value
            for field in fields(self)
            if (alias := field.metadata.get("alias")) is not None
            and (value := getattr(self, field.name)) is not None
            and alias.startswith(SDM_PREFIX)
        }

    @property
    def raw_data(self) -> dict[str, Any]:
        """Return raw data for the object."""
        result: dict[str, Any] = {}
        for k, v in self.to_dict(by_alias=True, omit_none=True).items():
            if k.startswith(SDM_PREFIX):
                if "traits" not in result:
                    result["traits"] = {}
                result["traits"][k] = v
            else:
                result[k] = v
        return result

    class Config(BaseConfig):
        code_generation_options = [
            "TO_DICT_ADD_BY_ALIAS_FLAG",
            "TO_DICT_ADD_OMIT_NONE_FLAG",
        ]
        serialize_by_alias = True



from __future__ import annotations

import asyncio
import json
from typing import Any, Awaitable, Callable, Dict, Optional
from unittest.mock import create_autospec, Mock, patch

import aiohttp
import pytest
from google.auth.exceptions import RefreshError, GoogleAuthError, TransportError
from google.api_core.exceptions import ClientError, Unauthenticated
from google.cloud import pubsub_v1
from google.oauth2.credentials import Credentials

from google_nest_sdm import diagnostics
from google_nest_sdm.auth import AbstractAuth
from google_nest_sdm.event import EventMessage
from google_nest_sdm.exceptions import (
    AuthException,
    ConfigurationException,
    SubscriberException,
)
from google_nest_sdm.google_nest_subscriber import (
    AbstractSubscriberFactory,
    GoogleNestSubscriber,
    get_api_env,
    refresh_creds,
)

from .conftest import DeviceHandler, EventCallback, StructureHandler, assert_diagnostics

PROJECT_ID = "project-id1"
SUBSCRIPTION_NAME = "projects/some-project-id/subscriptions/subscriber-id1"


class FakeSubscriberFactory(AbstractSubscriberFactory):
    def __init__(self) -> None:
        self.tasks: Optional[Any] = None

    async def async_create_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        topic_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        return

    async def async_delete_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        return

    async def async_new_subscriber(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
        async_callback: Callable[
            [pubsub_v1.subscriber.message.Message], Awaitable[None]
        ],
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        self._async_callback = async_callback
        if self.tasks:
            return asyncio.create_task(self.tasks.pop(0)())
        return None

    async def async_push_event(self, event: Dict[str, Any]) -> None:
        message = create_autospec(pubsub_v1.subscriber.message.Message, instance=True)
        message.data = json.dumps(event).encode()
        return await self._async_callback(message)


@pytest.fixture
def subscriber_factory() -> FakeSubscriberFactory:
    return FakeSubscriberFactory()


@pytest.fixture
def subscriber_client(
    subscriber_factory: FakeSubscriberFactory,
    auth_client: Callable[[], Awaitable[AbstractAuth]],
) -> Callable[[], Awaitable[GoogleNestSubscriber]]:
    async def make_subscriber(
        factory: Optional[AbstractSubscriberFactory] = subscriber_factory,
    ) -> GoogleNestSubscriber:
        auth = await auth_client()
        assert factory
        return GoogleNestSubscriber(auth, PROJECT_ID, SUBSCRIPTION_NAME, factory)

    return make_subscriber


async def test_refresh_creds() -> None:
    """Test low level refresh errors."""
    mock_refresh = Mock()
    mock_creds = Mock()
    mock_creds.refresh = mock_refresh
    refresh_creds(mock_creds)
    assert mock_refresh.call_count == 1


@pytest.mark.parametrize(
    ("raised", "expected"),
    [
        (RefreshError(), AuthException),
        (TransportError(), SubscriberException),
        (GoogleAuthError(), SubscriberException),
    ],
)
async def test_refresh_creds_error(raised: Exception, expected: Any) -> None:
    """Test low level refresh errors."""
    mock_refresh = Mock()
    mock_refresh.side_effect = raised
    mock_creds = Mock()
    mock_creds.refresh = mock_refresh
    with pytest.raises(expected):
        refresh_creds(mock_creds)


async def test_subscribe_no_events(
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[[], Awaitable[GoogleNestSubscriber]],
) -> None:
    device_id1 = device_handler.add_device(device_type="sdm.devices.types.device-type1")
    device_id2 = device_handler.add_device(device_type="sdm.devices.types.device-type2")
    structure_handler.add_structure()

    subscriber = await subscriber_client()
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    devices = device_manager.devices
    assert device_id1 in devices
    assert devices[device_id1].type == "sdm.devices.types.device-type1"
    assert device_id2 in devices
    assert devices[device_id2].type == "sdm.devices.types.device-type2"
    subscriber.stop_async()


async def test_subscribe_device_manager(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[[], Awaitable[GoogleNestSubscriber]],
) -> None:
    device_id1 = device_handler.add_device(device_type="sdm.devices.types.device-type1")
    device_id2 = device_handler.add_device(device_type="sdm.devices.types.device-type2")
    structure_handler.add_structure()

    subscriber = await subscriber_client()
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    devices = device_manager.devices
    assert device_id1 in devices
    assert devices[device_id1].type == "sdm.devices.types.device-type1"
    assert device_id2 in devices
    assert devices[device_id2].type == "sdm.devices.types.device-type2"
    subscriber.stop_async()

    assert_diagnostics(
        diagnostics.get_diagnostics(),
        {
            "subscriber": {
                "start": 1,
                "stop": 1,
            },
        },
    )


async def test_subscribe_update_trait(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[[], Awaitable[GoogleNestSubscriber]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.Connectivity": {
                "status": "ONLINE",
            },
        }
    )
    structure_handler.add_structure()

    subscriber = await subscriber_client()
    subscriber.cache_policy.event_cache_size = 5
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    devices = device_manager.devices
    assert device_id in devices
    device = devices[device_id]
    trait = device.traits["sdm.devices.traits.Connectivity"]
    assert "ONLINE" == trait.status

    event = {
        "eventId": "6f29332e-5537-47f6-a3f9-840c307340f5",
        "timestamp": "2020-10-10T07:09:06.851Z",
        "resourceUpdate": {
            "name": device_id,
            "traits": {
                "sdm.devices.traits.Connectivity": {
                    "status": "OFFLINE",
                }
            },
        },
        "userId": "AVPHwEv75jw4WFshx6-XhBLhotn3r8IXOzCusfSOn5QU",
    }
    await subscriber_factory.async_push_event(event)

    devices = device_manager.devices
    assert device_id in devices
    device = devices[device_id]
    trait = device.traits["sdm.devices.traits.Connectivity"]
    assert "OFFLINE" == trait.status
    subscriber.stop_async()

    assert_diagnostics(
        diagnostics.get_diagnostics(),
        {
            "subscriber": {
                "message_acked_count": 1,
                "message_received_count": 1,
                "start": 1,
                "stop": 1,
            },
        },
    )
    assert_diagnostics(
        device.get_diagnostics(),
        {
            "event_media": {"event": 1},
            "data": {
                "name": "**REDACTED**",
                "parentRelations": [],
                "traits": {"sdm.devices.traits.Connectivity": {"status": "OFFLINE"}},
                "type": "sdm.devices.types.device-type1",
            },
        },
    )


async def test_subscribe_device_manager_init(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[[], Awaitable[GoogleNestSubscriber]],
) -> None:
    device_id1 = device_handler.add_device(device_type="sdm.devices.types.device-type1")
    device_id2 = device_handler.add_device(device_type="sdm.devices.types.device-type2")
    structure_handler.add_structure()

    subscriber = await subscriber_client()
    start_async = subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    await start_async
    devices = device_manager.devices
    assert device_id1 in devices
    assert devices[device_id1].type == "sdm.devices.types.device-type1"
    assert device_id2 in devices
    assert devices[device_id2].type == "sdm.devices.types.device-type2"
    subscriber.stop_async()


async def test_subscriber_watchdog(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    auth_client: Callable[[], Awaitable[AbstractAuth]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    # Waits for the test to wake up the background thread
    event1 = asyncio.Event()

    async def task1() -> None:
        await event1.wait()

    event2 = asyncio.Event()

    async def task2() -> None:
        event2.set()

    subscriber_factory.tasks = [task1, task2]

    auth = await auth_client()
    subscriber = GoogleNestSubscriber(
        auth,
        PROJECT_ID,
        SUBSCRIPTION_NAME,
        subscriber_factory=subscriber_factory,
        watchdog_check_interval_seconds=0.1,
        watchdog_restart_delay_min_seconds=0.1,
    )
    assert len(subscriber_factory.tasks) == 2
    await subscriber.start_async()
    assert len(subscriber_factory.tasks) == 1
    # Wait for the subscriber to start, then shut it down
    event1.set()
    # Block until the new subscriber starts, and notifies this to wake up
    await event2.wait()
    assert len(subscriber_factory.tasks) == 0
    subscriber.stop_async()


async def test_subscriber_timeout(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[
        [Optional[AbstractSubscriberFactory]], Awaitable[GoogleNestSubscriber]
    ],
) -> None:
    class FailingFactory(FakeSubscriberFactory):
        async def async_new_subscriber(
            self,
            creds: Credentials,
            subscription_name: str,
            loop: asyncio.AbstractEventLoop,
            async_callback: Callable[
                [pubsub_v1.subscriber.message.Message], Awaitable[None]
            ],
        ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
            raise asyncio.TimeoutError("Some error")

    subscriber = await subscriber_client(FailingFactory())

    with pytest.raises(SubscriberException):
        await subscriber.start_async()
    subscriber.stop_async()

    assert_diagnostics(
        diagnostics.get_diagnostics(),
        {
            "subscriber": {
                "start": 1,
                "start.timeout_error": 1,
                "stop": 1,
            },
        },
    )
async def test_subscriber_error(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[
        [Optional[AbstractSubscriberFactory]], Awaitable[GoogleNestSubscriber]
    ],
) -> None:
    class FailingFactory(FakeSubscriberFactory):
        async def async_new_subscriber(
            self,
            creds: Credentials,
            subscription_name: str,
            loop: asyncio.AbstractEventLoop,
            async_callback: Callable[
                [pubsub_v1.subscriber.message.Message], Awaitable[None]
            ],
        ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
            raise ClientError("Some error")  # type: ignore

    subscriber = await subscriber_client(FailingFactory())

    with pytest.raises(SubscriberException):
        await subscriber.start_async()
    subscriber.stop_async()

    assert_diagnostics(
        diagnostics.get_diagnostics(),
        {
            "subscriber": {
                "start": 1,
                "start.api_error": 1,
                "stop": 1,
            },
        },
    )


async def test_subscriber_auth_error(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[
        [Optional[AbstractSubscriberFactory]], Awaitable[GoogleNestSubscriber]
    ],
) -> None:
    class FailingFactory(FakeSubscriberFactory):
        async def async_new_subscriber(
            self,
            creds: Credentials,
            subscription_name: str,
            loop: asyncio.AbstractEventLoop,
            async_callback: Callable[
                [pubsub_v1.subscriber.message.Message], Awaitable[None]
            ],
        ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
            raise Unauthenticated("Auth failure")  # type: ignore

    subscriber = await subscriber_client(FailingFactory())
    with pytest.raises(AuthException):
        await subscriber.start_async()
    subscriber.stop_async()


async def test_auth_refresh(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    refreshing_auth_client: Callable[[], Awaitable[AbstractAuth]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    async def auth_handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response({"token": "updated-token"})

    device_handler.token = "updated-token"
    device_id = device_handler.add_device(device_type="sdm.devices.types.device-type1")

    structure_handler.token = "updated-token"
    structure_handler.add_structure()

    app.router.add_get("/refresh-auth", auth_handler)

    auth = await refreshing_auth_client()
    subscriber = GoogleNestSubscriber(
        auth, PROJECT_ID, SUBSCRIPTION_NAME, subscriber_factory
    )
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    devices = device_manager.devices
    assert device_id in devices
    assert devices[device_id].type == "sdm.devices.types.device-type1"
    subscriber.stop_async()


async def test_auth_refresh_error(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    refreshing_auth_client: Callable[[], Awaitable[AbstractAuth]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    async def auth_handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(status=401)

    device_handler.token = "updated-token"
    device_handler.add_device()
    structure_handler.token = "updated-token"
    structure_handler.add_structure()

    app.router.add_get("/refresh-auth", auth_handler)

    auth = await refreshing_auth_client()
    subscriber = GoogleNestSubscriber(
        auth, PROJECT_ID, SUBSCRIPTION_NAME, subscriber_factory
    )

    with pytest.raises(AuthException):
        await subscriber.start_async()
    subscriber.stop_async()


async def test_subscriber_id_error(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    auth_client: Callable[[], Awaitable[AbstractAuth]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    auth = await auth_client()

    subscriber = GoogleNestSubscriber(
        auth, PROJECT_ID, "bad-subscriber-id", subscriber_factory
    )
    with pytest.raises(ConfigurationException):
        await subscriber.start_async()
    subscriber.stop_async()


async def test_subscribe_thread_update(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[[], Awaitable[GoogleNestSubscriber]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.CameraClipPreview": {},
            "sdm.devices.traits.CameraMotion": {},
        }
    )
    structure_handler.add_structure()

    subscriber = await subscriber_client()
    subscriber.cache_policy.event_cache_size = 5
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    devices = device_manager.devices
    assert device_id in devices

    subscriber_callback = EventCallback()
    subscriber.set_update_callback(subscriber_callback.async_handle_event)

    device_callback = EventCallback()
    devices[device_id].add_event_callback(device_callback.async_handle_event)

    event = {
        "eventId": "6f29332e-5537-47f6-a3f9-840c307340f5",
        "timestamp": "2020-10-10T07:09:06.851Z",
        "resourceUpdate": {
            "name": device_id,
            "events": {
                "sdm.devices.events.CameraMotion.Motion": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "eventId": "n:1",
                },
                "sdm.devices.events.CameraClipPreview.ClipPreview": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "previewUrl": "image-url-1",
                },
            },
        },
        "userId": "AVPHwEv75jw4WFshx6-XhBLhotn3r8IXOzCusfSOn5QU",
        "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
        "resourcegroup": [
            "enterprises/project-id1/devices/device-id1",
        ],
        "eventThreadState": "STARTED",
    }
    await subscriber_factory.async_push_event(event)

    # Verify the message is received. The full content is verified below.
    assert len(subscriber_callback.messages) == 1
    message: EventMessage = subscriber_callback.messages[0]
    assert message.event_id == "6f29332e-5537-47f6-a3f9-840c307340f5"

    # Device-level callback also received the message
    assert len(device_callback.messages) == 1

    # End the thread (resource update is identical)
    event = {
        "eventId": "7f29332e-5537-47f6-a3f9-840c307340f5",
        "timestamp": "2020-10-10T07:09:07.851Z",
        "resourceUpdate": {
            "name": device_id,
            "events": {
                "sdm.devices.events.CameraMotion.Motion": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "eventId": "n:1",
                },
                "sdm.devices.events.CameraClipPreview.ClipPreview": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "previewUrl": "image-url-1",
                },
            },
        },
        "userId": "AVPHwEv75jw4WFshx6-XhBLhotn3r8IXOzCusfSOn5QU",
        "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
        "resourcegroup": [
            "enterprises/project-id1/devices/device-id1",
        ],
        "eventThreadState": "ENDED",
    }
    await subscriber_factory.async_push_event(event)

    assert len(subscriber_callback.messages) == 1
    message = subscriber_callback.messages[0]
    assert message.event_id == "6f29332e-5537-47f6-a3f9-840c307340f5"
    assert message.event_sessions
    assert len(message.event_sessions) == 1
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." in message.event_sessions
    events = message.event_sessions["CjY5Y3VKaTZwR3o4Y19YbTVfMF..."]
    assert len(events) == 2
    assert "sdm.devices.events.CameraMotion.Motion" in events
    assert "sdm.devices.events.CameraClipPreview.ClipPreview" in events

    # Device-level callback invoked with both raw messages
    assert len(device_callback.messages) == 2
    message = device_callback.messages[0]
    assert message.event_id == "6f29332e-5537-47f6-a3f9-840c307340f5"
    message = device_callback.messages[1]
    assert message.event_id == "7f29332e-5537-47f6-a3f9-840c307340f5"

    subscriber.stop_async()


async def test_subscribe_thread_update_new_events(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[[], Awaitable[GoogleNestSubscriber]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.CameraClipPreview": {},
            "sdm.devices.traits.CameraMotion": {},
            "sdm.devices.traits.CameraPerson": {},
        }
    )
    structure_handler.add_structure()

    subscriber = await subscriber_client()
    subscriber.cache_policy.event_cache_size = 5
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    devices = device_manager.devices
    assert device_id in devices

    callback = EventCallback()
    subscriber.set_update_callback(callback.async_handle_event)

    device_callback = EventCallback()
    devices[device_id].add_event_callback(device_callback.async_handle_event)

    event = {
        "eventId": "6f29332e-5537-47f6-a3f9-840c307340f5",
        "timestamp": "2020-10-10T07:09:06.851Z",
        "resourceUpdate": {
            "name": device_id,
            "events": {
                "sdm.devices.events.CameraMotion.Motion": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "eventId": "n:1",
                },
                "sdm.devices.events.CameraClipPreview.ClipPreview": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "previewUrl": "image-url-1",
                },
            },
        },
        "userId": "AVPHwEv75jw4WFshx6-XhBLhotn3r8IXOzCusfSOn5QU",
        "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
        "resourcegroup": [
            "enterprises/project-id1/devices/device-id1",
        ],
        "eventThreadState": "STARTED",
    }
    await subscriber_factory.async_push_event(event)

    # End the thread (resource update is identical)
    event2 = {
        "eventId": "7f29332e-5537-47f6-a3f9-840c307340f5",
        "timestamp": "2020-10-10T07:09:07.851Z",
        "resourceUpdate": {
            "name": device_id,
            "events": {
                "sdm.devices.events.CameraMotion.Motion": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "eventId": "n:1",
                },
                "sdm.devices.events.CameraPerson.Person": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "eventId": "n:2",
                },
                "sdm.devices.events.CameraClipPreview.ClipPreview": {
                    "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                    "previewUrl": "image-url-1",
                },
            },
        },
        "userId": "AVPHwEv75jw4WFshx6-XhBLhotn3r8IXOzCusfSOn5QU",
        "eventThreadId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
        "resourcegroup": [
            "enterprises/project-id1/devices/device-id1",
        ],
        "eventThreadState": "ENDED",
    }
    await subscriber_factory.async_push_event(event2)

    assert len(callback.messages) == 2
    message: EventMessage = callback.messages[0]
    assert message.event_id == "6f29332e-5537-47f6-a3f9-840c307340f5"
    assert message.event_sessions
    assert len(message.event_sessions) == 1
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." in message.event_sessions
    events = message.event_sessions["CjY5Y3VKaTZwR3o4Y19YbTVfMF..."]
    assert len(events) == 2
    assert "sdm.devices.events.CameraMotion.Motion" in events
    assert "sdm.devices.events.CameraClipPreview.ClipPreview" in events

    message = callback.messages[1]
    assert message.event_id == "7f29332e-5537-47f6-a3f9-840c307340f5"
    assert message.event_sessions
    assert len(message.event_sessions) == 1
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." in message.event_sessions
    events = message.event_sessions["CjY5Y3VKaTZwR3o4Y19YbTVfMF..."]
    assert len(events) == 1
    assert "sdm.devices.events.CameraPerson.Person" in events

    # Device also receives the same messages
    assert len(device_callback.messages) == 2

    subscriber.stop_async()


async def test_message_ack_timeout(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_factory: FakeSubscriberFactory,
    subscriber_client: Callable[
        [Optional[AbstractSubscriberFactory]], Awaitable[GoogleNestSubscriber]
    ],
    fake_event_message: Callable[[Dict[str, Any]], EventMessage],
) -> None:

    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.Connectivity": {
                "status": "ONLINE",
            },
        }
    )
    structure_id = structure_handler.add_structure()

    subscriber = await subscriber_client(subscriber_factory)
    subscriber.cache_policy.event_cache_size = 5
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()
    devices = device_manager.devices
    assert device_id in devices

    async def async_handle_event(_: Any) -> None:
        await asyncio.sleep(10)

    subscriber.set_update_callback(async_handle_event)
    event = {
        "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
        "timestamp": "2019-01-01T00:00:01Z",
        "relationUpdate": {
            "type": "CREATED",
            "subject": structure_id,
            "object": device_id,
        },
        "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
    }
    with patch(
        "google_nest_sdm.google_nest_subscriber.MESSAGE_ACK_TIMEOUT_SECONDS", 0.01
    ):
        with pytest.raises(TimeoutError, match="Message ack timeout"):
            await subscriber_factory.async_push_event(event)

    subscriber.stop_async()


async def test_refresh_hack_on_invalid_thermostat_traits(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    subscriber_client: Callable[[], Awaitable[GoogleNestSubscriber]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    """Test that when invalid thermostat traits are ignored and triggers a refresh."""

    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.ThermostatEco": {
                "availableModes": ["MANUAL_ECO", "OFF"],
                "mode": "MANUAL_ECO",
                "heatCelsius": 20.0,
                "coolCelsius": 22.0,
            },
            "sdm.devices.traits.ThermostatHvac": {
                "status": "HEATING",
            },
            "sdm.devices.traits.ThermostatMode": {
                "availableModes": ["HEAT", "COOL", "HEATCOOL", "OFF"],
                "mode": "HEAT",
            },
            "sdm.devices.traits.ThermostatTemperatureSetpoint": {
                "heatCelsius": 20.0,
            },
        }
    )
    structure_handler.add_structure()

    subscriber = await subscriber_client()
    subscriber.cache_policy.event_cache_size = 5
    await subscriber.start_async()
    device_manager = await subscriber.async_get_device_manager()

    device = device_manager.devices.get(device_id)
    assert device

    # Verify initial device state
    trait = device.traits.get("sdm.devices.traits.ThermostatEco")
    assert trait
    assert set(trait.available_modes) == set(["MANUAL_ECO", "OFF"])
    assert trait.mode == "MANUAL_ECO"
    assert trait.heat_celsius == 20.0
    assert trait.cool_celsius == 22.0

    trait = device.traits.get("sdm.devices.traits.ThermostatMode")
    assert trait
    assert set(trait.available_modes) == set(["HEAT", "COOL", "HEATCOOL", "OFF"])
    assert trait.mode == "HEAT"

    trait = device.traits.get("sdm.devices.traits.ThermostatHvac")
    assert trait
    assert trait.status == "HEATING"

    trait = device.traits.get("sdm.devices.traits.ThermostatTemperatureSetpoint")
    assert trait
    assert trait.heat_celsius == 20.0

    # Update the state on the server to something arbitrary. Later we will verify
    # that a request was made to get the latest server state.
    trait = device_handler.devices[0]["traits"]["sdm.devices.traits.ThermostatEco"]
    trait["heatCelsius"] = 19.0

    # Simulate a case where the nest publisher sends an invalid message. This
    # will be ignored and will trigger another state refresh.
    await subscriber_factory.async_push_event(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": device.name,
                "traits": {
                    "sdm.devices.traits.ThermostatMode": {
                        "mode": "OFF",
                        "availableModes": ["OFF"],
                    },
                    "sdm.devices.traits.ThermostatEco": {
                        "availableModes": ["OFF", "MANUAL_ECO"],
                        "mode": "OFF",
                        "heatCelsius": 0.0,
                        "coolCelsius": 0.0,
                    },
                    "sdm.devices.traits.ThermostatTemperatureSetpoint": {},
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )

    device = device_manager.devices.get(device_id)
    assert device

    # Verify the invalid message is dropped and the server update is reflected
    trait = device.traits.get("sdm.devices.traits.ThermostatEco")
    assert trait
    assert set(trait.available_modes) == set(["MANUAL_ECO", "OFF"])
    assert trait.mode == "MANUAL_ECO"
    assert trait.heat_celsius == 19.0  # State update reflected
    assert trait.cool_celsius == 22.0

    trait = device.traits.get("sdm.devices.traits.ThermostatMode")
    assert trait
    assert set(trait.available_modes) == set(["HEAT", "COOL", "HEATCOOL", "OFF"])
    assert trait.mode == "HEAT"

    trait = device.traits.get("sdm.devices.traits.ThermostatHvac")
    assert trait
    assert trait.status == "HEATING"

    trait = device.traits.get("sdm.devices.traits.ThermostatTemperatureSetpoint")
    assert trait
    assert trait.heat_celsius == 20.0


def test_api_env_prod() -> None:
    env = get_api_env("prod")
    assert (
        env.authorize_url_format
        == "https://nestservices.google.com/partnerconnections/{project_id}/auth"
    )
    assert env.api_url == "https://smartdevicemanagement.googleapis.com/v1"

    env = get_api_env(None)
    assert (
        env.authorize_url_format
        == "https://nestservices.google.com/partnerconnections/{project_id}/auth"
    )
    assert env.api_url == "https://smartdevicemanagement.googleapis.com/v1"


def test_api_env_preprod() -> None:
    env = get_api_env("preprod")
    assert env.authorize_url_format == (
        "https://sdmresourcepicker-preprod.sandbox.google.com/partnerconnections/"
        "{project_id}/auth"
    )
    assert env.api_url == "https://preprod-smartdevicemanagement.googleapis.com/v1"


def test_api_env_invalid() -> None:
    with pytest.raises(ValueError):
        get_api_env("invalid")



async def test_topic_id(
    app: aiohttp.web.Application,
    device_handler: DeviceHandler,
    structure_handler: StructureHandler,
    auth_client: Callable[[], Awaitable[AbstractAuth]],
    subscriber_factory: FakeSubscriberFactory,
) -> None:
    """Test that creates a client with a specific topic id."""

    auth = await auth_client()
    # Just exercise the constructor
    subscriber = GoogleNestSubscriber(
        auth,
        PROJECT_ID,
        subscriber_name=SUBSCRIPTION_NAME,
        topic_name="projects/some-project-id/topics/some-topic-id",
        subscriber_factory=subscriber_factory,
        watchdog_check_interval_seconds=0.1,
        watchdog_restart_delay_min_seconds=0.1,
    )
    await subscriber.create_subscription()



"""Events from pubsub subscriber."""

from __future__ import annotations

from abc import ABC, abstractmethod
import base64
import binascii
from dataclasses import dataclass, field
import datetime
from enum import StrEnum
import hashlib
import json
import logging
import traceback
from typing import Any, Iterable, Mapping, ClassVar

from mashumaro import DataClassDictMixin, field_options
from mashumaro.config import (
    BaseConfig,
)
from mashumaro.types import SerializationStrategy

from .auth import AbstractAuth
from .exceptions import DecodeException
from .registry import Registry

__all__ = [
    "EventMessage",
    "CameraMotionEvent",
    "CameraPersonEvent",
    "CameraSoundEvent",
    "DoorbellChimeEvent",
    "CameraClipPreviewEvent",
    "EventImageType",
    "EventProcessingError",
]

EVENT_ID = "eventId"
EVENT_SESSION_ID = "eventSessionId"
TIMESTAMP = "timestamp"
RESOURCE_UPDATE = "resourceUpdate"
NAME = "name"
TRAITS = "traits"
EVENTS = "events"
PREVIEW_URL = "previewUrl"
ZONES = "zones"
EVENT_THREAD_STATE_ENDED = "ENDED"

# Event images expire 30 seconds after the event is published
EVENT_IMAGE_EXPIRE_SECS = 30

# Camera clip previews don't list an expiration in the API. Lets say 15 minutes
# as an arbitrary number for now.
CAMERA_CLIP_PREVIEW_EXPIRE_SECS = 15 * 60

EVENT_MAP = Registry()

_LOGGER = logging.getLogger(__name__)


class EventType(StrEnum):
    """Types of events."""

    CAMERA_MOTION = "sdm.devices.events.CameraMotion.Motion"
    CAMERA_PERSON = "sdm.devices.events.CameraPerson.Person"
    CAMERA_SOUND = "sdm.devices.events.CameraSound.Sound"
    DOORBELL_CHIME = "sdm.devices.events.DoorbellChime.Chime"
    CAMERA_CLIP_PREVIEW = "sdm.devices.events.CameraClipPreview.ClipPreview"


class EventProcessingError(Exception):
    """Raised when there was an error handling an event."""


@dataclass(frozen=True)
class EventImageContentType(DataClassDictMixin):
    """Event image content type."""

    content_type: str

    def __str__(self) -> str:
        """Return a string representation of the event image type."""
        return self.content_type


class EventImageType(ABC):
    IMAGE = EventImageContentType("image/jpeg")
    CLIP_PREVIEW = EventImageContentType("video/mp4")
    IMAGE_PREVIEW = EventImageContentType("image/gif")

    @staticmethod
    def from_string(content_type: str) -> EventImageContentType:
        """Parse an EventImageType from a string representation."""
        if content_type == EventImageType.CLIP_PREVIEW.content_type:
            return EventImageType.CLIP_PREVIEW
        elif content_type == EventImageType.IMAGE.content_type:
            return EventImageType.IMAGE
        elif content_type == EventImageType.IMAGE_PREVIEW.content_type:
            return EventImageType.IMAGE_PREVIEW
        else:
            return EventImageContentType(content_type)


@dataclass
class EventToken:
    """Identifier for a unique event."""

    event_session_id: str = field(metadata=field_options(alias="eventSessionId"))
    event_id: str = field(metadata=field_options(alias="eventId"))

    def encode(self) -> str:
        """Encode the event token as a serialized string."""
        data = [self.event_session_id, self.event_id]
        b = json.dumps(data).encode("utf-8")
        return base64.b64encode(b).decode("utf-8")

    @staticmethod
    def decode(content: str) -> EventToken:
        """Decode an event token into a class."""
        try:
            s = base64.b64decode(content).decode("utf-8")
        except binascii.Error as err:
            raise DecodeException from err
        data = json.loads(s)
        if not isinstance(data, list) or len(data) != 2:
            raise DecodeException("Unexpected data type: %s", data)
        return EventToken(data[0], data[1])

    def __repr__(self) -> str:
        if not self.event_id:
            return "<EventToken event_session_id" + self.event_session_id + ">"
        return (
            "<EventToken event_session_id"
            + self.event_session_id
            + " event_id="
            + self.event_id
            + ">"
        )


class EventImageTypeSerializationStrategy(SerializationStrategy):

    def serialize(self, value: EventImageContentType) -> str:
        return value.content_type

    def deserialize(self, value: str) -> EventImageContentType:
        return EventImageType.from_string(value)


@dataclass
class ImageEventBase(DataClassDictMixin, ABC):
    """Base class for all image related event types."""

    event_session_id: str = field(metadata=field_options(alias="eventSessionId"))
    """ID used to associate separate messages with a single event."""

    timestamp: datetime.datetime
    """Timestamp when the event occurred."""

    event_id: str = field(metadata=field_options(alias="eventId"), default="")
    """ID used to associate separate messages with a single event."""

    event_image_type: EventImageContentType = field(default=EventImageType.IMAGE)
    """Type of the event."""

    zones: list[str] = field(default_factory=list)
    """List of zones for the event."""

    @property
    def event_token(self) -> str:
        """An identifier of this session / event combination."""
        token = EventToken(self.event_session_id, self.event_id)
        return token.encode()

    @property
    @abstractmethod
    def event_type(self) -> EventType:
        """The type of event."""

    @property
    def expires_at(self) -> datetime.datetime:
        """Timestamp when the message expires."""
        return self.timestamp + datetime.timedelta(seconds=EVENT_IMAGE_EXPIRE_SECS)

    @property
    def is_expired(self) -> bool:
        """Return true if the event expiration has passed."""
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        return self.expires_at < now

    def as_dict(self) -> dict[str, Any]:
        """Return as a dict form that can be serialized for persistence."""
        return {
            "event_type": self.event_type,
            "event_data": self.to_dict(),
            "timestamp": self.timestamp.isoformat(),
            "event_image_type": str(self.event_image_type),
        }

    @staticmethod
    def parse_event_dict(data: dict[str, Any]) -> ImageEventBase | None:
        """Parse from a persisted serialized dictionary."""
        event_type = data["event_type"]
        event_data = data["event_data"]
        event = _BuildEvent(event_type, event_data)
        if event and "event_image_type" in data:
            event.event_image_type = EventImageType.from_string(
                data["event_image_type"]
            )
        return event

    class Config(BaseConfig):
        serialization_strategy = {
            EventImageContentType: EventImageTypeSerializationStrategy(),
        }
        code_generation_options = [
            "TO_DICT_ADD_BY_ALIAS_FLAG",
            "TO_DICT_ADD_OMIT_NONE_FLAG",
        ]
        allow_deserialization_not_by_alias = True


@EVENT_MAP.register()
@dataclass
class CameraMotionEvent(ImageEventBase):
    """Motion has been detected by the camera."""

    NAME: ClassVar[EventType] = EventType.CAMERA_MOTION
    event_image_type: EventImageContentType = field(default=EventImageType.IMAGE)

    @property
    def event_type(self) -> EventType:
        """The type of event."""
        return EventType.CAMERA_MOTION


@EVENT_MAP.register()
@dataclass
class CameraPersonEvent(ImageEventBase):
    """A person has been detected by the camera."""

    NAME: ClassVar[EventType] = EventType.CAMERA_PERSON
    event_image_type: EventImageContentType = field(default=EventImageType.IMAGE)

    @property
    def event_type(self) -> EventType:
        """The type of event."""
        return EventType.CAMERA_PERSON


@EVENT_MAP.register()
@dataclass
class CameraSoundEvent(ImageEventBase):
    """Sound has been detected by the camera."""

    NAME: ClassVar[EventType] = EventType.CAMERA_SOUND
    event_image_type: EventImageContentType = field(default=EventImageType.IMAGE)

    @property
    def event_type(self) -> EventType:
        """The type of event."""
        return EventType.CAMERA_SOUND


@EVENT_MAP.register()
@dataclass
class DoorbellChimeEvent(ImageEventBase):
    """The doorbell has been pressed."""

    NAME: ClassVar[EventType] = EventType.DOORBELL_CHIME
    event_image_type: EventImageContentType = field(default=EventImageType.IMAGE)

    @property
    def event_type(self) -> EventType:
        """The type of event."""
        return EventType.DOORBELL_CHIME


@EVENT_MAP.register()
@dataclass
class CameraClipPreviewEvent(ImageEventBase):
    """A video clip is available for preview, without extra download."""

    NAME: ClassVar[EventType] = EventType.CAMERA_CLIP_PREVIEW
    event_image_type: EventImageContentType = field(default=EventImageType.CLIP_PREVIEW)

    preview_url: str = field(metadata=field_options(alias="previewUrl"), default="")
    """A url 10 second frame video file in mp4 format."""

    @property
    def event_type(self) -> EventType:
        """The type of event."""
        return EventType.CAMERA_CLIP_PREVIEW

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Validate the event id to use a URL hash as the event id.

        Since clip preview events already have a url associated with them,
        we don't have an event id for downloading the image.
        """
        if not (preview_url := d.get("previewUrl", d.get("preview_url"))):
            raise ValueError("missing required field previewUrl")
        d["eventId"] = hashlib.blake2b(preview_url.encode()).hexdigest()
        return d

    @property
    def expires_at(self) -> datetime.datetime:
        """Event ids do not expire."""
        return self.timestamp + datetime.timedelta(
            seconds=CAMERA_CLIP_PREVIEW_EXPIRE_SECS
        )


@dataclass
class RelationUpdate(DataClassDictMixin):
    """Represents a relational update for a resource."""

    type: str
    """Type of relation event 'CREATED', 'UPDATED', 'DELETED'."""

    subject: str
    """Resource that the object is now in relation with."""

    object: str
    """Resource that triggered the event."""


def _BuildEvent(
    event_type: str, event_data: Mapping[str, Any]
) -> ImageEventBase | None:
    if event_type not in EVENT_MAP:
        _LOGGER.debug("Event type %s not found (%s)", event_type, EVENT_MAP.keys())
        return None
    cls = EVENT_MAP[event_type]
    try:
        return cls.from_dict(event_data)  # type: ignore
    except Exception as err:
        traceback.print_exc()
        _LOGGER.debug("Failed to parse event: %s (event_data=%s)", err, event_data)
        raise err


def session_event_image_type(events: Iterable[ImageEventBase]) -> EventImageContentType:
    """Determine the event type to use based on the events in the session."""
    for event in events:
        if event.event_image_type != EventImageType.IMAGE:
            return event.event_image_type
    return EventImageType.IMAGE


class UpdateEventsSerializationStrategy(SerializationStrategy, use_annotations=True):
    """Parser to ignore invalid parent relations."""

    def serialize(self, value: dict[str, ImageEventBase]) -> dict[str, Any]:
        return {k: v.to_dict(by_alias=True) for k, v in value.items()}

    def deserialize(self, value: dict[str, Any]) -> dict[str, ImageEventBase]:
        result = {}
        for event_type, event_data in value.items():
            image_event = _BuildEvent(event_type, event_data)
            if not image_event:
                continue
            result[event_type] = image_event
        return result


@dataclass
class EventMessage(DataClassDictMixin):
    """Event for a change in trait value or device action."""

    timestamp: datetime.datetime
    event_id: str = field(metadata=field_options(alias="eventId"))
    resource_update_name: str | None = field(default=None)
    resource_update_events: dict[str, ImageEventBase] | None = field(default=None)
    resource_update_traits: dict[str, Any] | None = field(default=None)
    event_thread_state: str | None = field(
        metadata=field_options(alias="eventThreadState"), default=None
    )
    relation_update: RelationUpdate | None = field(
        metadata=field_options(alias="relationUpdate"), default=None
    )

    _auth: AbstractAuth = field(init=False, metadata={"serialize": "omit"})

    @classmethod
    def create_event(
        cls, raw_data: dict[str, Any], auth: AbstractAuth
    ) -> "EventMessage":
        """Initialize an EventMessage."""
        event_data = {**raw_data}
        _LOGGER.debug("EventMessage raw_data=%s", event_data)
        if update := event_data.get(RESOURCE_UPDATE):
            if name := update.get(NAME):
                event_data["resource_update_name"] = name
            if events := update.get(EVENTS):
                timestamp = event_data.get(TIMESTAMP)
                for event_updates in events.values():
                    event_updates[TIMESTAMP] = timestamp
                event_data["resource_update_events"] = events
            if traits := update.get(TRAITS):
                event_data["resource_update_traits"] = traits
                event_data["resource_update_traits"][NAME] = update.get(NAME)

        event = cls.from_dict(event_data)
        event._auth = auth
        return event

    @property
    def event_sessions(self) -> dict[str, dict[str, ImageEventBase]] | None:
        events = self.resource_update_events
        if not events:
            return None
        event_sessions: dict[str, dict[str, ImageEventBase]] = {}
        for event_name, event in events.items():
            d = event_sessions.get(event.event_session_id, {})
            d[event_name] = event
            event_sessions[event.event_session_id] = d
        # Build associations between all events
        for event_session_id, event_dict in event_sessions.items():
            event_image_type = session_event_image_type(events.values())
            for event_type, event in event_dict.items():
                event.event_image_type = event_image_type
        return event_sessions

    @property
    def raw_data(self) -> dict[str, Any]:
        """Return raw data for the event."""
        return self.to_dict(by_alias=True)

    def with_events(
        self,
        event_keys: Iterable[str],
        merge_data: dict[str, ImageEventBase] | None = None,
    ) -> EventMessage:
        """Create a new EventMessage minus some existing events by key."""
        new_message = EventMessage.create_event(self.to_dict(by_alias=True), self._auth)
        if not merge_data:
            merge_data = {}
        new_events = {}
        for key in event_keys:
            if (
                new_message.resource_update_events
                and key in new_message.resource_update_events
            ):
                new_events[key] = new_message.resource_update_events[key]
            elif merge_data and key in merge_data:
                new_events[key] = merge_data[key]
        new_message.resource_update_events = new_events
        return new_message

    @property
    def is_thread_ended(self) -> bool:
        """Return true if the message indicates the thread is ended."""
        return self.event_thread_state == EVENT_THREAD_STATE_ENDED

    def __repr__(self) -> str:
        """Debug information."""
        return f"EventMessage{self.to_dict()}"

    class Config(BaseConfig):
        serialization_strategy = {
            dict[str, ImageEventBase]: UpdateEventsSerializationStrategy(),
        }
        code_generation_options = [
            "TO_DICT_ADD_BY_ALIAS_FLAG",
            "TO_DICT_ADD_OMIT_NONE_FLAG",
        ]



from google_nest_sdm.structure import Structure


def test_no_traits() -> None:
    raw = {
        "name": "my/structure/name",
    }
    structure = Structure.MakeStructure(raw)
    assert "my/structure/name" == structure.name
    assert "sdm.structures.traits.Info" not in structure.traits


def test_empty_traits() -> None:
    raw = {
        "name": "my/structure/name",
        "traits": {},
    }
    structure = Structure.MakeStructure(raw)
    assert "my/structure/name" == structure.name
    assert "sdm.structures.traits.Info" not in structure.traits


def test_info_traits() -> None:
    raw = {
        "name": "my/structure/name",
        "traits": {
            "sdm.structures.traits.Info": {
                "customName": "Structure Name",
            },
        },
    }
    structure = Structure.MakeStructure(raw)
    assert "my/structure/name" == structure.name
    assert "sdm.structures.traits.Info" in structure.traits
    trait = structure.traits["sdm.structures.traits.Info"]
    assert "Structure Name" == trait.custom_name


def test_room_info_traits() -> None:
    raw = {
        "name": "my/structure/name",
        "traits": {
            "sdm.structures.traits.RoomInfo": {
                "customName": "Structure Name",
            },
        },
    }
    structure = Structure.MakeStructure(raw)
    assert "my/structure/name" == structure.name
    assert "sdm.structures.traits.RoomInfo" in structure.traits
    trait = structure.traits["sdm.structures.traits.RoomInfo"]
    assert "Structure Name" == trait.custom_name



import asyncio
from unittest.mock import Mock, patch

import pytest

from google_nest_sdm.exceptions import TranscodeException
from google_nest_sdm.transcoder import Transcoder

BINARY = "/bin/true"


async def test_transcoder_file_not_exist(tmp_path: str) -> None:
    t = Transcoder(BINARY, path_prefix=tmp_path)
    with pytest.raises(TranscodeException):
        await t.transcode_clip("in_file.mp4", "out_file.gif")


async def test_transcoder_output_already_exists(tmp_path: str) -> None:
    t = Transcoder(BINARY, path_prefix=tmp_path)
    with open(f"{tmp_path}/in_file.mp4", mode="w") as f:
        f.write("some-input")
    with open(f"{tmp_path}/out_file.gif", mode="w") as f:
        f.write("some-output")
    with pytest.raises(TranscodeException):
        await t.transcode_clip("in_file.mp4", "out_file.gif")


async def test_transcoder(tmp_path: str) -> None:
    t = Transcoder(BINARY, path_prefix=tmp_path)
    with open(f"{tmp_path}/in_file.mp4", mode="w") as f:
        f.write("some-input")
    with patch(
        "google_nest_sdm.transcoder.asyncio.create_subprocess_shell"
    ) as mock_shell:
        process_mock = Mock()
        future: asyncio.Future = asyncio.Future()
        future.set_result(("", ""))
        process_mock.communicate.return_value = future
        process_mock.returncode = 0
        mock_shell.return_value = process_mock
        await t.transcode_clip("in_file.mp4", "out_file.gif")


async def test_transcoder_failure(tmp_path: str) -> None:
    t = Transcoder("/bin/false", path_prefix=tmp_path)
    with open(f"{tmp_path}/in_file.mp4", mode="w") as f:
        f.write("some-input")
    with patch(
        "google_nest_sdm.transcoder.asyncio.create_subprocess_shell"
    ) as mock_shell, pytest.raises(TranscodeException):
        process_mock = Mock()
        future: asyncio.Future = asyncio.Future()
        future.set_result(("", ""))
        process_mock.communicate.return_value = future
        process_mock.returncode = 1
        mock_shell.return_value = process_mock
        await t.transcode_clip("in_file.mp4", "out_file.gif")



"""Library for traits about devices."""

import datetime
from typing import Any, Dict, ClassVar
from dataclasses import dataclass, field

import aiohttp
from mashumaro import field_options, DataClassDictMixin

from .traits import CommandDataClass, TraitType


@dataclass
class ConnectivityTrait(DataClassDictMixin):
    """This trait belongs to any device that has connectivity information."""

    NAME: ClassVar[TraitType] = TraitType.CONNECTIVITY

    status: str
    """Device connectivity status.

    Return:
        "OFFLINE", "ONLINE"
    """


@dataclass
class FanTrait(DataClassDictMixin, CommandDataClass):
    """This trait belongs to any device that can control the fan."""

    NAME: ClassVar[TraitType] = TraitType.FAN

    timer_mode: str | None = field(
        metadata=field_options(alias="timerMode"), default=None
    )
    """Timer mode for the fan.

    Return:
        "ON", "OFF"
    """

    timer_timeout: datetime.datetime | None = field(
        metadata=field_options(alias="timerTimeout"), default=None
    )

    async def set_timer(
        self, timer_mode: str, duration: int | None = None
    ) -> aiohttp.ClientResponse:
        """Change the fan timer."""
        data: Dict[str, Any] = {
            "command": "sdm.devices.commands.Fan.SetTimer",
            "params": {
                "timerMode": timer_mode,
            },
        }
        if duration:
            data["params"]["duration"] = f"{duration}s"
        return await self.cmd.execute(data)


@dataclass
class InfoTrait(DataClassDictMixin):
    """This trait belongs to any device for device-related information."""

    NAME: ClassVar[TraitType] = TraitType.INFO

    custom_name: str | None = field(
        metadata=field_options(alias="customName"), default=None
    )
    """Name of the device."""


@dataclass
class HumidityTrait(DataClassDictMixin):
    """This trait belongs to any device that has a sensor to measure humidity."""

    NAME: ClassVar[TraitType] = TraitType.HUMIDITY

    ambient_humidity_percent: float = field(
        metadata=field_options(alias="ambientHumidityPercent")
    )
    """Percent humidity, measured at the device."""


@dataclass
class TemperatureTrait(DataClassDictMixin):
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME: ClassVar[TraitType] = TraitType.TEMPERATURE

    ambient_temperature_celsius: float = field(
        metadata=field_options(alias="ambientTemperatureCelsius")
    )
    """Percent humidity, measured at the device."""



"""Library with functions for manipulating WebRTC requests/responses."""

from enum import StrEnum


class SDPDirection(StrEnum):
    """SDP direction constants."""

    SENDRECV = "sendrecv"
    SENDONLY = "sendonly"
    RECVONLY = "recvonly"
    INACTIVE = "inactive"


class SDPMediaKind(StrEnum):
    """SDP media kind constants."""

    AUDIO = "audio"
    VIDEO = "video"
    APPLICATION = "application"


def _get_media_direction(sdp: str, kind: SDPMediaKind) -> SDPDirection | None:
    """Retrieves the direction of media tracks from the SDP based on the kind (audio/video)."""

    # Track if we are in the desired media section
    in_media_section = False

    for line in sdp.split("\r\n"):
        # Check if the line is a media description line
        if line.startswith("m="):
            in_media_section = line.startswith(f"m={kind}")
        # If we're in the desired media section, check for direction
        if in_media_section and line.startswith("a="):
            for direction in SDPDirection:
                if line.startswith(f"a={direction}"):
                    return direction
    return None


def _update_direction_in_answer(
    answer_sdp: str,
    kind: SDPMediaKind,
    old_direction: SDPDirection,
    new_direction: SDPDirection,
) -> str:
    """Updates the direction of a specific media track in the SDP answer if it matches a certain direction."""

    # Update the SDP
    updated_sdp_lines = []
    in_media_section = False
    for line in answer_sdp.split("\r\n"):
        if line.startswith("m="):
            in_media_section = line.startswith(f"m={kind}")
        if in_media_section and line.startswith("a="):
            # Update the direction line if it matches the kind
            if line.startswith(f"a={old_direction}"):
                updated_sdp_lines.append(
                    line.replace(f"a={old_direction}", f"a={new_direction}")
                )
                continue
        updated_sdp_lines.append(line)
    return "\r\n".join(updated_sdp_lines)


def _add_foundation_to_candidates(sdp: str) -> str:
    """Adds a foundation value to all ICE candidates in the SDP if it does not already exist."""

    updated_sdp_lines = []
    index = 1
    for line in sdp.split("\r\n"):
        if line.startswith("a=candidate: "):
            updated_sdp_lines.append(
                line.replace("a=candidate: ", f"a=candidate:{index} ")
            )
            index += 1
            continue
        updated_sdp_lines.append(line)
    return "\r\n".join(updated_sdp_lines)


def fix_mozilla_sdp_answer(offer_sdp: str, answer_sdp: str) -> str:
    """Fix the answer SDP which is rejected by Firefox.

    1. If offer SDP is recvonly, the direction of answer SDP must not be sendrecv.
    2. If the ICE candidates in answer SDP must contain "foundation" field.
    """
    if "mozilla" in offer_sdp:
        if (
            _get_media_direction(sdp=offer_sdp, kind=SDPMediaKind.VIDEO)
            == SDPDirection.RECVONLY
        ):
            answer_sdp = _update_direction_in_answer(
                answer_sdp=answer_sdp,
                kind=SDPMediaKind.VIDEO,
                old_direction=SDPDirection.SENDRECV,
                new_direction=SDPDirection.SENDONLY,
            )
        if (
            _get_media_direction(sdp=offer_sdp, kind=SDPMediaKind.AUDIO)
            == SDPDirection.RECVONLY
        ):
            answer_sdp = _update_direction_in_answer(
                answer_sdp=answer_sdp,
                kind=SDPMediaKind.AUDIO,
                old_direction=SDPDirection.SENDRECV,
                new_direction=SDPDirection.SENDONLY,
            )
        return _add_foundation_to_candidates(answer_sdp)
    return answer_sdp



"""Tests for thermostat traits."""

from typing import Any, Awaitable, Callable, Dict

import aiohttp
import pytest

from google_nest_sdm import google_nest_api
from google_nest_sdm.device import Device

from .conftest import DeviceHandler, NewHandler, Recorder


def test_thermostat_eco_traits(fake_device: Callable[[Dict[str, Any]], Device]) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {
                "sdm.devices.traits.ThermostatEco": {
                    "availableModes": ["MANUAL_ECHO", "OFF"],
                    "mode": "MANUAL_ECHO",
                    "heatCelsius": 20.0,
                    "coolCelsius": 22.0,
                },
            },
        }
    )
    assert "sdm.devices.traits.ThermostatEco" in device.traits
    trait = device.traits["sdm.devices.traits.ThermostatEco"]
    assert ["MANUAL_ECHO", "OFF"] == trait.available_modes
    assert "MANUAL_ECHO" == trait.mode
    assert 20.0 == trait.heat_celsius
    assert 22.0 == trait.cool_celsius


def test_thermostat_hvac_traits(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {
                "sdm.devices.traits.ThermostatHvac": {
                    "status": "HEATING",
                },
            },
        }
    )
    assert "sdm.devices.traits.ThermostatHvac" in device.traits
    trait = device.traits["sdm.devices.traits.ThermostatHvac"]
    assert "HEATING" == trait.status


def test_thermostat_mode_traits(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {
                "sdm.devices.traits.ThermostatMode": {
                    "availableModes": ["HEAT", "COOL", "HEATCOOL", "OFF"],
                    "mode": "COOL",
                },
            },
        }
    )
    assert "sdm.devices.traits.ThermostatMode" in device.traits
    trait = device.traits["sdm.devices.traits.ThermostatMode"]
    assert ["HEAT", "COOL", "HEATCOOL", "OFF"] == trait.available_modes
    assert "COOL" == trait.mode


def test_thermostat_temperature_setpoint_traits(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {
                "sdm.devices.traits.ThermostatTemperatureSetpoint": {
                    "heatCelsius": 20.0,
                    "coolCelsius": 22.0,
                },
            },
        }
    )
    assert "sdm.devices.traits.ThermostatTemperatureSetpoint" in device.traits
    trait = device.traits["sdm.devices.traits.ThermostatTemperatureSetpoint"]
    assert 20.0 == trait.heat_celsius
    assert 22.0 == trait.cool_celsius


@pytest.mark.parametrize(
    "data",
    [
        ({}),
        ({"heatCelsius": 20.0}),
        ({"coolCelsius": 22.0}),
        ({"heatCelsius": 20.0, "coolCelsius": 22.0}),
    ],
)
def test_thermostat_temperature_setpoint_optional_fields(
    fake_device: Callable[[Dict[str, Any]], Device], data: dict[str, Any]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {"sdm.devices.traits.ThermostatTemperatureSetpoint": data},
        }
    )
    assert "sdm.devices.traits.ThermostatTemperatureSetpoint" in device.traits
    assert device.thermostat_temperature_setpoint


def test_thermostat_multiple_traits(
    fake_device: Callable[[Dict[str, Any]], Device]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {
                "sdm.devices.traits.ThermostatEco": {
                    "availableModes": ["MANUAL_ECHO", "OFF"],
                    "mode": "MANUAL_ECHO",
                    "heatCelsius": 21.0,
                    "coolCelsius": 22.0,
                },
                "sdm.devices.traits.ThermostatHvac": {
                    "status": "HEATING",
                },
                "sdm.devices.traits.ThermostatMode": {
                    "availableModes": ["HEAT", "COOL", "HEATCOOL", "OFF"],
                    "mode": "COOL",
                },
                "sdm.devices.traits.ThermostatTemperatureSetpoint": {
                    "heatCelsius": 23.0,
                    "coolCelsius": 24.0,
                },
            },
        }
    )
    assert "sdm.devices.traits.ThermostatEco" in device.traits
    assert "sdm.devices.traits.ThermostatHvac" in device.traits
    assert "sdm.devices.traits.ThermostatMode" in device.traits
    assert "sdm.devices.traits.ThermostatTemperatureSetpoint" in device.traits
    trait = device.traits["sdm.devices.traits.ThermostatEco"]
    assert ["MANUAL_ECHO", "OFF"] == trait.available_modes
    assert "MANUAL_ECHO" == trait.mode
    assert 21.0 == trait.heat_celsius
    assert 22.0 == trait.cool_celsius
    trait = device.traits["sdm.devices.traits.ThermostatHvac"]
    assert "HEATING" == trait.status
    trait = device.traits["sdm.devices.traits.ThermostatMode"]
    assert ["HEAT", "COOL", "HEATCOOL", "OFF"] == trait.available_modes
    assert "COOL" == trait.mode
    trait = device.traits["sdm.devices.traits.ThermostatTemperatureSetpoint"]
    assert 23.0 == trait.heat_celsius
    assert 24.0 == trait.cool_celsius


@pytest.mark.parametrize(
    "data",
    [
        ({}),
        ({"mode": "OFF"}),
    ],
)
def test_thermostat_eco_optional_fields(
    fake_device: Callable[[Dict[str, Any]], Device], data: dict[str, Any]
) -> None:
    device = fake_device(
        {
            "name": "my/device/name",
            "traits": {"sdm.devices.traits.ThermostatEco": data},
        }
    )
    assert "sdm.devices.traits.ThermostatEco" in device.traits
    assert device.thermostat_eco
    assert device.thermostat_eco.mode == "OFF"


async def test_fan_set_timer(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.Fan": {
                "timerMode": "OFF",
            },
        }
    )
    post_handler = NewHandler(recorder, [{}])
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device_id == device.name
    trait = device.traits["sdm.devices.traits.Fan"]
    assert trait.timer_mode == "OFF"
    await trait.set_timer("ON", 3600)
    assert recorder.request == {
        "command": "sdm.devices.commands.Fan.SetTimer",
        "params": {
            "timerMode": "ON",
            "duration": "3600s",
        },
    }


async def test_thermostat_eco_set_mode(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.ThermostatEco": {
                "availableModes": ["MANUAL_ECO", "OFF"],
                "mode": "MANUAL_ECO",
                "heatCelsius": 20.0,
                "coolCelsius": 22.0,
            },
        }
    )
    post_handler = NewHandler(recorder, [{}])
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.ThermostatEco"]
    assert trait.mode == "MANUAL_ECO"
    await trait.set_mode("OFF")
    assert recorder.request == {
        "command": "sdm.devices.commands.ThermostatEco.SetMode",
        "params": {"mode": "OFF"},
    }


async def test_thermostat_mode_set_mode(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.ThermostatMode": {
                "availableModes": ["HEAT", "COOL", "HEATCOOL", "OFF"],
                "mode": "COOL",
            },
        }
    )
    post_handler = NewHandler(recorder, [{}])
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.ThermostatMode"]
    assert trait.mode == "COOL"
    await trait.set_mode("HEAT")
    assert recorder.request == {
        "command": "sdm.devices.commands.ThermostatMode.SetMode",
        "params": {"mode": "HEAT"},
    }


async def test_thermostat_temperature_set_point(
    app: aiohttp.web.Application,
    recorder: Recorder,
    device_handler: DeviceHandler,
    api_client: Callable[[], Awaitable[google_nest_api.GoogleNestAPI]],
) -> None:
    device_id = device_handler.add_device(
        traits={
            "sdm.devices.traits.ThermostatTemperatureSetpoint": {
                "heatCelsius": 23.0,
                "coolCelsius": 24.0,
            },
        }
    )
    post_handler = NewHandler(recorder, [{}, {}, {}])
    app.router.add_post(f"/{device_id}:executeCommand", post_handler)

    api = await api_client()
    devices = await api.async_get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.name == device_id
    trait = device.traits["sdm.devices.traits.ThermostatTemperatureSetpoint"]
    assert trait.heat_celsius == 23.0
    assert trait.cool_celsius == 24.0
    await trait.set_heat(25.0)
    assert recorder.request == {
        "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetHeat",
        "params": {"heatCelsius": 25.0},
    }

    await trait.set_cool(26.0)
    assert recorder.request == {
        "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetCool",
        "params": {"coolCelsius": 26.0},
    }

    await trait.set_range(27.0, 28.0)
    assert recorder.request == {
        "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetRange",
        "params": {
            "heatCelsius": 27.0,
            "coolCelsius": 28.0,
        },
    }



"""Admin Client library for the Google Nest SDM API.

This manages administrative tasks for setting up pubsub topics and subscriptions.

This library exists to provide an asyncio interface given that the current pubsub
clients are synchronous.
"""

import logging
import re
import asyncio
from typing import Any
from dataclasses import dataclass, field

from .diagnostics import SUBSCRIBER_DIAGNOSTICS as DIAGNOSTICS
from .auth import AbstractAuth
from .exceptions import (
    ApiException,
    NotFoundException,
    ApiForbiddenException,
    ConfigurationException,
)

_LOGGER = logging.getLogger(__name__)

__all__ = [
    "AdminClient",
    "EligibleTopics",
    "EligibleSubscriptions",
    "validate_subscription_name",
    "validate_topic_name",
    "PUBSUB_API_HOST",
]

PUBSUB_API_HOST = "https://pubsub.googleapis.com/v1"
SDM_MANAGED_TOPIC_FORMAT = (
    "projects/sdm-prod/topics/enterprise-{device_access_project_id}"
)

# Used to catch invalid subscriber id
EXPECTED_SUBSCRIBER_REGEXP = re.compile("^projects/[^/]+/subscriptions/[^/]+$")

# Used to catch a topic misconfiguration
EXPECTED_TOPIC_REGEXP = re.compile("^projects/[^/]+/topics/[^/]+$")

# Topic prefix for the project
EXPECTED_PROJECS_PREFIX = re.compile("^projects/[^/]+$")


@dataclass
class EligibleTopics:
    """Eligible topics for the project."""

    topic_names: list[str] = field(default_factory=list)


@dataclass
class EligibleSubscriptions:
    """Eligible topics for the project."""

    subscription_names: list[str] = field(default_factory=list)


def validate_subscription_name(subscription_name: str) -> None:
    """Validates that a subscription name is correct.

    Raises ConfigurationException on failure.
    """
    if not EXPECTED_SUBSCRIBER_REGEXP.match(subscription_name):
        DIAGNOSTICS.increment("subscription_name_invalid")
        _LOGGER.debug("Subscription name did not match pattern: %s", subscription_name)
        raise ConfigurationException(
            "Subscription misconfigured. Expected subscriber_id to "
            f"match '{EXPECTED_SUBSCRIBER_REGEXP.pattern}' but was "
            f"'{subscription_name}'"
        )


def validate_topic_name(topic_name: str) -> None:
    """Validates that a topic name is correct.

    Raises ConfigurationException on failure.
    """
    if not EXPECTED_TOPIC_REGEXP.match(topic_name):
        DIAGNOSTICS.increment("topic_name_invalid")
        _LOGGER.debug("Topic name did not match pattern: %s", topic_name)
        raise ConfigurationException(
            "Subscription misconfigured. Expected topic name to "
            f"match '{EXPECTED_TOPIC_REGEXP.pattern}' but was "
            f"'{topic_name}'."
        )


def validate_projects_prefix(project_path: str) -> None:
    """Validates that a topic or subscription prefix is correct.

    Raises ConfigurationException on failure.
    """
    if not EXPECTED_PROJECS_PREFIX.match(project_path):
        DIAGNOSTICS.increment("topic_prefix_invalid")
        _LOGGER.debug("Topic prefix did not match pattern: %s", project_path)
        raise ConfigurationException(
            "Subscription misconfigured. Expected topic name to "
            f"match '{EXPECTED_PROJECS_PREFIX.pattern}' but was "
            f"'{project_path}'."
        )


class AdminClient:
    """Admin client for the Google Nest SDM API."""

    def __init__(
        self,
        auth: AbstractAuth,
        cloud_project_id: str,
    ) -> None:
        """Initialize the admin client.

        The auth instance must be configured with the correct host (PUBSUB_API_HOST).
        """
        self._cloud_project_id = cloud_project_id
        self._auth = auth

    async def create_topic(self, topic_name: str) -> None:
        """Create a pubsub topic for the project."""
        validate_topic_name(topic_name)
        await self._auth.request("put", topic_name)

    async def delete_topic(self, topic_name: str) -> None:
        """Delete a pubsub topic for the project."""
        validate_topic_name(topic_name)
        await self._auth.request("delete", topic_name)

    async def list_topics(self, projects_prefix: str) -> list[str]:
        """List the pubsub topics for the project.

        The topic prefix should be in the format `projects/{console_project_id}`.
        """
        validate_projects_prefix(projects_prefix)
        response = await self._auth.get_json(f"{projects_prefix}/topics")
        return [topic["name"] for topic in response["topics"]]

    async def get_topic(self, topic_name: str) -> dict[str, Any]:
        """Get a pubsub topic for the project."""
        validate_topic_name(topic_name)
        return await self._auth.get_json(topic_name)

    async def create_subscription(
        self, topic_name: str, subscription_name: str
    ) -> None:
        """Create a pubsub subscription for the project."""
        validate_topic_name(topic_name)
        validate_subscription_name(subscription_name)
        body = {"topic": topic_name}
        await self._auth.request("put", subscription_name, json=body)

    async def delete_subscription(self, subscription_name: str) -> None:
        """Delete a pubsub subscription for the project."""
        validate_subscription_name(subscription_name)
        await self._auth.request("delete", subscription_name)

    async def list_subscriptions(self, projects_prefix: str) -> list[dict[str, Any]]:
        """List the pubsub subscriptions for the project.
        The projects_prefix should be in the format `projects/{console_project_id}`.
        """
        validate_projects_prefix(projects_prefix)
        response = await self._auth.get_json(f"{projects_prefix}/subscriptions")
        return response["subscriptions"]  # type: ignore[no-any-return]

    async def list_eligible_topics(
        self, device_access_project_id: str
    ) -> EligibleTopics:
        """List the eligible topics for the project.

        This will try to find any topics already created for the project by either
        the device access console or by the user.
        """

        sdm_topic_name = SDM_MANAGED_TOPIC_FORMAT.format(
            device_access_project_id=device_access_project_id
        )

        async def get_sdm_topic() -> str | None:
            try:
                await self.get_topic(sdm_topic_name)
            except ApiForbiddenException:
                _LOGGER.debug(
                    "SDM topic exists but we do not have permission to access it (expected)"
                )
                # The SDM topic exists. It is normal that we do not have permission
                # to access it.
                return sdm_topic_name
            except NotFoundException:
                _LOGGER.debug(
                    "SDM topic does not exist, proceeding to check cloud projects"
                )
                return None
            except ApiException as err:
                _LOGGER.error(
                    "Unexpected error retrieving an SDM created topic: %s", err
                )
                raise ApiException("Error retrieving SDM created topic") from err
            _LOGGER.debug(
                "SDM topic exists and we have permission to access it (unexpected)"
            )
            return sdm_topic_name

        async def get_cloud_topics() -> list[str]:
            try:
                return await self.list_topics(f"projects/{self._cloud_project_id}")
            except ApiException as err:
                _LOGGER.error("Unexpected error listing topics: %s", err)
                raise ApiException(
                    "Error while listing existing cloud console topics"
                ) from err

        (sdm_topic_task, cloud_topics_task) = await asyncio.gather(
            get_sdm_topic(), get_cloud_topics()
        )
        topics = []
        if sdm_topic_task:
            topics.append(sdm_topic_task)
        topics.extend(cloud_topics_task)
        return EligibleTopics(topic_names=topics)

    async def list_eligible_subscriptions(
        self, expected_topic_name: str
    ) -> EligibleSubscriptions:
        """Return a set of eligible subscriptions for the project."""
        subscriptions = await self.list_subscriptions(
            f"projects/{self._cloud_project_id}"
        )
        return EligibleSubscriptions(
            subscription_names=[
                sub["name"]
                for sub in subscriptions
                if sub["topic"] == expected_topic_name
            ]
        )



"""Library for exceptions using the Google Nest SDM API and subscriber."""


class GoogleNestException(Exception):
    """Base class for all client exceptions."""


class SubscriberException(GoogleNestException):
    """Raised during problems subscribing to events and updates."""


class ApiException(GoogleNestException):
    """Raised during problems talking to the API."""


class AuthException(ApiException):
    """Raised due to auth problems talking to API or subscriber."""


class NotFoundException(ApiException):
    """Raised when the API returns an error that a resource was not found."""


class ApiForbiddenException(ApiException):
    """Raised when the user is not authorized to perform a specific function."""


class ConfigurationException(GoogleNestException):
    """Raised due to misconfiguration problems."""


class DecodeException(GoogleNestException):
    """Raised when failing to decode a token."""


class TranscodeException(GoogleNestException):
    """Raised when failing to transcode media."""



"""Tests for the request client library."""

from typing import Awaitable, Callable

import aiohttp
import pytest
from aiohttp.test_utils import TestClient, TestServer
from yarl import URL

from google_nest_sdm.auth import AbstractAuth
from google_nest_sdm.exceptions import ApiException


async def test_request(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        assert request.path == "/path-prefix/some-path"
        assert request.headers["header-1"] == "value-1"
        assert request.headers["Authorization"] == "Bearer some-token"
        assert request.query == {"client_id": "some-client-id"}
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    app.router.add_get("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    resp = await auth.request(
        "get",
        "some-path",
        headers={"header-1": "value-1"},
        params={"client_id": "some-client-id"},
    )
    resp.raise_for_status()
    data = await resp.json()
    assert data == {"some-key": "some-value"}


async def test_auth_header(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    """Test that a request with an Ahthorization header is preserved."""

    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        assert request.path == "/path-prefix/some-path"
        assert request.headers["header-1"] == "value-1"
        assert request.headers["Authorization"] == "Basic other-token"
        assert request.query["client_id"] == "some-client-id"
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    app.router.add_get("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    resp = await auth.request(
        "get",
        "some-path",
        headers={"header-1": "value-1", "Authorization": "Basic other-token"},
        params={"client_id": "some-client-id"},
    )
    resp.raise_for_status()
    data = await resp.json()
    assert data == {"some-key": "some-value"}


async def test_full_url(
    app: aiohttp.web.Application,
    client: Callable[[], Awaitable[TestClient]],
    server: Callable[[], Awaitable[TestServer]],
    auth_client: Callable[[str], Awaitable[AbstractAuth]],
) -> None:
    """Test that a request with an Ahthorization header is preserved."""

    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        assert request.path == "/path-prefix/some-path"
        assert request.headers["header-1"] == "value-1"
        assert request.headers["Authorization"] == "Bearer some-token"
        assert request.query["client_id"] == "some-client-id"
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    app.router.add_get("/path-prefix/some-path", handler)

    test_server = await server()

    def client_make_url(url: str) -> URL:
        assert url == "https://example/path-prefix/some-path"
        return test_server.make_url("/path-prefix/some-path")

    test_client = await client()
    test_client.make_url = client_make_url  # type: ignore

    auth = await auth_client("/path-prefix")
    resp = await auth.request(
        "get",
        "https://example/path-prefix/some-path",
        headers={"header-1": "value-1"},
        params={"client_id": "some-client-id"},
    )
    resp.raise_for_status()
    data = await resp.json()
    assert data == {"some-key": "some-value"}


async def test_get_json_response(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        assert request.query["client_id"] == "some-client-id"
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    app.router.add_get("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    data = await auth.get_json("some-path", params={"client_id": "some-client-id"})
    assert data == {"some-key": "some-value"}


async def test_get_json_response_unexpected(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response(["value1", "value2"])

    app.router.add_get("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    with pytest.raises(ApiException):
        await auth.get_json("some-path")


async def test_get_json_response_unexpected_text(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(text="body")

    app.router.add_get("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    with pytest.raises(ApiException):
        await auth.get_json("some-path")


async def test_post_json_response(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        body = await request.json()
        assert body == {"client_id": "some-client-id"}
        return aiohttp.web.json_response(
            {
                "some-key": "some-value",
            }
        )

    app.router.add_post("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    data = await auth.post_json("some-path", json={"client_id": "some-client-id"})
    assert data == {"some-key": "some-value"}


async def test_post_json_response_unexpected(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.json_response(["value1", "value2"])

    app.router.add_post("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    with pytest.raises(ApiException):
        await auth.post_json("some-path")


async def test_post_json_response_unexpected_text(
    app: aiohttp.web.Application, auth_client: Callable[[str], Awaitable[AbstractAuth]]
) -> None:
    async def handler(request: aiohttp.web.Request) -> aiohttp.web.Response:
        return aiohttp.web.Response(text="body")

    app.router.add_post("/path-prefix/some-path", handler)

    auth = await auth_client("/path-prefix")
    with pytest.raises(ApiException):
        await auth.post_json("some-path")



"""Authentication library, implemented by users of the API.

This library is a simple `aiohttp` that handles authentication when talking
to the API. Users are expected to provide their own implementation that provides
credentials obtained using the standard Google authentication approaches
described at https://developers.google.com/nest/device-access/api/authorization

An implementation of `AbstractAuth` implements `async_get_access_token`
to provide authentication credentials to the SDM library. The implementation is
responsible for managing the lifecycle of the token (any persistence needed,
or refresh to deal with expiration, etc).
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from asyncio import TimeoutError
from typing import Any
from http import HTTPStatus

import aiohttp
from aiohttp.client_exceptions import ClientError
from google.auth.credentials import Credentials
from google.oauth2.credentials import Credentials as OAuthCredentials
from mashumaro.mixins.json import DataClassJSONMixin

from .exceptions import (
    ApiException,
    AuthException,
    ApiForbiddenException,
    NotFoundException,
)

_LOGGER = logging.getLogger(__name__)

__all__ = ["AbstractAuth"]

HTTP_UNAUTHORIZED = 401
AUTHORIZATION_HEADER = "Authorization"
ERROR = "error"
STATUS = "status"
MESSAGE = "message"


@dataclass
class Status(DataClassJSONMixin):
    """Status of the media item."""

    code: int = field(default=HTTPStatus.OK)
    """The status code, which should be an enum value of google.rpc.Code"""

    message: str | None = None
    """A developer-facing error message, which should be in English"""

    details: list[dict[str, Any]] = field(default_factory=list)
    """A list of messages that carry the error details"""


@dataclass
class Error:
    """Error details from the API response."""

    status: str | None = None
    code: int | None = None
    message: str | None = None
    details: list[dict[str, Any]] | None = field(default_factory=list)

    def __str__(self) -> str:
        """Return a string representation of the error details."""
        error_message = ""
        if self.status:
            error_message += self.status
        if self.code:
            if error_message:
                error_message += f" ({self.code})"
            else:
                error_message += str(self.code)
        if self.message:
            if error_message:
                error_message += ": "
            error_message += self.message
        if self.details:
            error_message += f"\nError details: ({self.details})"
        return error_message


@dataclass
class ErrorResponse(DataClassJSONMixin):
    """A response message that contains an error message."""

    error: Error | None = None


class AbstractAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(self, websession: aiohttp.ClientSession, host: str):
        """Initialize the AbstractAuth."""
        self._websession = websession
        self._host = host

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def async_get_creds(self) -> Credentials:
        """Return creds for subscriber API."""
        token = await self.async_get_access_token()
        return OAuthCredentials(token=token)

    async def request(
        self,
        method: str,
        url: str,
        **kwargs: Any,
    ) -> aiohttp.ClientResponse:
        """Make a request."""
        headers = kwargs.get("headers")

        if headers is None:
            headers = {}
        else:
            headers = dict(headers)
            del kwargs["headers"]
        if AUTHORIZATION_HEADER not in headers:
            try:
                access_token = await self.async_get_access_token()
            except TimeoutError as err:
                raise ApiException(f"Timeout requesting API token: {err}") from err
            except ClientError as err:
                raise AuthException(f"Access token failure: {err}") from err
            headers[AUTHORIZATION_HEADER] = f"Bearer {access_token}"
        if not (url.startswith("http://") or url.startswith("https://")):
            url = f"{self._host}/{url}"
        _LOGGER.debug("request[%s]=%s", method, url)
        if method == "post" and "json" in kwargs:
            _LOGGER.debug("request[post json]=%s", kwargs["json"])
        try:
            return await self._request(method, url, headers=headers, **kwargs)
        except (ClientError, TimeoutError) as err:
            raise ApiException(f"Error connecting to API: {err}") from err

    async def _request(
        self, method: str, url: str, headers: dict[str, str], **kwargs: Any
    ) -> aiohttp.ClientResponse:
        response = await self._websession.request(method, url, **kwargs, headers=headers)
        return await AbstractAuth._raise_for_status(response)


    async def get(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        """Make a get request."""
        return await self.request("get", url, **kwargs)

    async def get_json(self, url: str, **kwargs: Any) -> dict[str, Any]:
        """Make a get request and return json response."""
        resp = await self.get(url, **kwargs)
        try:
            result = await resp.json()
        except ClientError as err:
            raise ApiException("Server returned malformed response") from err
        if not isinstance(result, dict):
            raise ApiException("Server return malformed response: %s" % result)
        _LOGGER.debug("response=%s", result)
        return result

    async def post(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        """Make a post request."""
        return await self.request("post", url, **kwargs)

    async def post_json(self, url: str, **kwargs: Any) -> dict[str, Any]:
        """Make a post request and return a json response."""
        resp = await self.post(url, **kwargs)
        try:
            result = await resp.json()
        except ClientError as err:
            raise ApiException("Server returned malformed response") from err
        if not isinstance(result, dict):
            raise ApiException("Server returned malformed response: %s" % result)
        _LOGGER.debug("response=%s", result)
        return result

    @classmethod
    async def _raise_for_status(
        cls, resp: aiohttp.ClientResponse
    ) -> aiohttp.ClientResponse:
        """Raise exceptions on failure methods."""
        error_detail = await cls._error_detail(resp)
        try:
            resp.raise_for_status()
        except aiohttp.ClientResponseError as err:
            error_message = f"{err.message} response from API ({resp.status})"
            if error_detail:
                error_message += f": {error_detail}"
            if err.status == HTTPStatus.FORBIDDEN:
                raise ApiForbiddenException(error_message)
            if err.status == HTTPStatus.UNAUTHORIZED:
                raise AuthException(error_message)
            if err.status == HTTPStatus.NOT_FOUND:
                raise NotFoundException(error_message)
            raise ApiException(error_message) from err
        except aiohttp.ClientError as err:
            raise ApiException(f"Error from API: {err}") from err
        return resp

    @classmethod
    async def _error_detail(cls, resp: aiohttp.ClientResponse) -> Error | None:
        """Returns an error message string from the APi response."""
        if resp.status < 400:
            return None
        try:
            result = await resp.text()
        except ClientError:
            return None
        try:
            error_response = ErrorResponse.from_json(result)
        except (LookupError, ValueError):
            return None
        return error_response.error



"""Traits belonging to doorbell devices."""

from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import ClassVar

from .event import DoorbellChimeEvent, EventType
from .traits import TraitType

_LOGGER = logging.getLogger(__name__)


@dataclass
class DoorbellChimeTrait:
    """For any device that supports a doorbell chime and related press events."""

    NAME: ClassVar[TraitType] = TraitType.DOORBELL_CHIME
    EVENT_NAME: ClassVar[EventType] = DoorbellChimeEvent.NAME



"""Traits for thermostats."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Final, ClassVar

import aiohttp
from mashumaro import field_options, DataClassDictMixin

from .traits import CommandDataClass, TraitType

__all__ = [
    "ThermostatEcoTrait",
    "ThermostatHvacTrait",
    "ThermostatModeTrait",
    "ThermostatTemperatureSetpointTrait",
]

STATUS: Final = "status"
AVAILABLE_MODES: Final = "availableModes"
MODE: Final = "mode"


@dataclass
class ThermostatEcoTrait(DataClassDictMixin, CommandDataClass):
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME: ClassVar[TraitType] = TraitType.THERMOSTAT_ECO

    available_modes: list[str] = field(
        metadata=field_options(alias="availableModes"), default_factory=list
    )
    """List of supported Eco modes."""

    mode: str = field(metadata=field_options(alias="mode"), default="OFF")
    """Eco mode of the thermostat."""

    heat_celsius: float | None = field(
        metadata=field_options(alias="heatCelsius"), default=None
    )
    """Lowest temperature where thermostat begins heating."""

    cool_celsius: float | None = field(
        metadata=field_options(alias="coolCelsius"), default=None
    )
    """Highest cooling temperature where thermostat begins cooling."""

    async def set_mode(self, mode: str) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatEco.SetMode",
            "params": {"mode": mode},
        }
        return await self.cmd.execute(data)


@dataclass
class ThermostatHvacTrait:
    """This trait belongs to devices that can report HVAC details."""

    NAME: ClassVar[TraitType] = TraitType.THERMOSTAT_HVAC

    status: str
    """HVAC status of the thermostat."""


@dataclass
class ThermostatModeTrait(DataClassDictMixin, CommandDataClass):
    """This trait belongs to devices that support different thermostat modes."""

    NAME: ClassVar[TraitType] = TraitType.THERMOSTAT_MODE

    available_modes: list[str] = field(metadata=field_options(alias="availableModes"))
    """List of supported thermostat modes."""

    mode: str = field(metadata=field_options(alias="mode"))
    """Mode of the thermostat."""

    async def set_mode(self, mode: str) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatMode.SetMode",
            "params": {"mode": mode},
        }
        return await self.cmd.execute(data)


@dataclass
class ThermostatTemperatureSetpointTrait(DataClassDictMixin, CommandDataClass):
    """This trait belongs to devices that support setting target temperature."""

    NAME: ClassVar[TraitType] = TraitType.THERMOSTAT_TEMPERATURE_SETPOINT

    heat_celsius: float | None = field(
        metadata=field_options(alias="heatCelsius"), default=None
    )
    """Lowest temperature where thermostat begins heating."""

    cool_celsius: float | None = field(
        metadata=field_options(alias="coolCelsius"), default=None
    )
    """Highest cooling temperature where thermostat begins cooling."""

    async def set_heat(self, heat: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetHeat",
            "params": {"heatCelsius": heat},
        }
        return await self.cmd.execute(data)

    async def set_cool(self, cool: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetCool",
            "params": {"coolCelsius": cool},
        }
        return await self.cmd.execute(data)

    async def set_range(self, heat: float, cool: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetRange",
            "params": {
                "heatCelsius": heat,
                "coolCelsius": cool,
            },
        }
        return await self.cmd.execute(data)



"""Traits for structures / rooms."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Mapping
from mashumaro import field_options

from .model import TraitDataClass


@dataclass
class InfoTrait:
    """This trait belongs to any structure for structure-related information."""

    custom_name: str | None = field(
        metadata=field_options(alias="customName"), default=None
    )
    """Name of the structure."""


@dataclass
class RoomInfoTrait:
    """This trait belongs to any structure for room-related information."""

    custom_name: str = field(metadata=field_options(alias="customName"))
    """Name of the structure."""


@dataclass
class Structure(TraitDataClass):
    """Class that represents a structure object in the Google Nest SDM API."""

    name: str
    """Resource name of the structure e.g. 'enterprises/XYZ/structures/123'."""

    info: InfoTrait | None = field(
        metadata=field_options(alias="sdm.structures.traits.Info"), default=None
    )
    room_info: RoomInfoTrait | None = field(
        metadata=field_options(alias="sdm.structures.traits.RoomInfo"), default=None
    )

    @classmethod
    def MakeStructure(cls, raw_data: Mapping[str, Any]) -> Structure:
        """Create a structure with the appropriate traits."""
        return cls.parse_trait_object(raw_data)



from __future__ import annotations

import base64
import datetime
import json
from typing import Any, Callable, Dict, Optional

import pytest

from google_nest_sdm.event import (
    CameraClipPreviewEvent,
    EventImageType,
    EventMessage,
    EventToken,
    ImageEventBase,
)
from google_nest_sdm.exceptions import DecodeException


def test_camera_sound_event(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraSound.Sound": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "0120ecc7-3b57-4eb4-9941-91609f189fb4" == event.event_id
    ts = datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)

    assert ts == event.timestamp
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name
    events = event.resource_update_events
    assert events is not None
    assert "sdm.devices.events.CameraSound.Sound" in events
    e = events["sdm.devices.events.CameraSound.Sound"]
    assert "FWWVQVUdGNUlTU2V4MGV2aTNXV..." == e.event_id
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id
    assert ts == e.timestamp
    expire_ts = datetime.datetime(2019, 1, 1, 0, 0, 31, tzinfo=datetime.timezone.utc)
    assert expire_ts == e.expires_at


def test_camera_person_event(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraPerson.Person": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "0120ecc7-3b57-4eb4-9941-91609f189fb4" == event.event_id
    assert (
        datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)
        == event.timestamp
    )
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name
    events = event.resource_update_events
    assert events is not None
    assert "sdm.devices.events.CameraPerson.Person" in events
    e = events["sdm.devices.events.CameraPerson.Person"]
    assert "FWWVQVUdGNUlTU2V4MGV2aTNXV..." == e.event_id
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id


def test_camera_motion_event(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraMotion.Motion": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "0120ecc7-3b57-4eb4-9941-91609f189fb4" == event.event_id
    assert (
        datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)
        == event.timestamp
    )
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name
    events = event.resource_update_events
    assert events is not None
    assert "sdm.devices.events.CameraMotion.Motion" in events
    e = events["sdm.devices.events.CameraMotion.Motion"]
    assert "FWWVQVUdGNUlTU2V4MGV2aTNXV..." == e.event_id
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id


def test_doorbell_chime_event(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.DoorbellChime.Chime": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "0120ecc7-3b57-4eb4-9941-91609f189fb4" == event.event_id
    assert (
        datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)
        == event.timestamp
    )
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name
    events = event.resource_update_events
    assert events is not None
    assert "sdm.devices.events.DoorbellChime.Chime" in events
    e = events["sdm.devices.events.DoorbellChime.Chime"]
    assert "FWWVQVUdGNUlTU2V4MGV2aTNXV..." == e.event_id
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id


def test_relation(fake_event_message: Callable[[Dict[str, Any]], EventMessage]) -> None:
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "relationUpdate": {
                "type": "CREATED",
                "subject": "enterprises/project-id/structures/structure-id",
                "object": "enterprises/project-id/devices/device-id",
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "0120ecc7-3b57-4eb4-9941-91609f189fb4" == event.event_id
    assert (
        datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)
        == event.timestamp
    )
    assert event.resource_update_name is None
    assert event.resource_update_events is None
    assert event.resource_update_traits is None
    update = event.relation_update
    assert update is not None
    assert "CREATED" == update.type
    assert "enterprises/project-id/structures/structure-id" == update.subject
    assert "enterprises/project-id/devices/device-id" == update.object


def test_camera_clip_preview_event(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "201fcd21-967a-4f82-8082-5073bd09d31f",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraClipPreview.ClipPreview": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "previewUrl": "https://previewUrl/...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "201fcd21-967a-4f82-8082-5073bd09d31f" == event.event_id
    ts = datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)

    assert ts == event.timestamp
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name

    events: Optional[Dict[str, ImageEventBase]] = event.resource_update_events
    assert events is not None
    assert "sdm.devices.events.CameraClipPreview.ClipPreview" in events
    e = events["sdm.devices.events.CameraClipPreview.ClipPreview"]
    assert isinstance(e, CameraClipPreviewEvent)
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id
    assert ts == e.timestamp
    assert (
        "1cd28a21db0705a03f612d1df956444094a8c85a75fd35dfd277a81b5a9ad2a8c18"
        "2a9a7decb2936664d6e1344915850a31fc3c828c3813765d1c73be1e958f3" == e.event_id
    )
    assert "https://previewUrl/..." == e.preview_url
    expire_ts = datetime.datetime(2019, 1, 1, 0, 15, 1, tzinfo=datetime.timezone.utc)
    assert expire_ts == e.expires_at


def test_event_serialization(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "201fcd21-967a-4f82-8082-5073bd09d31f",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraClipPreview.ClipPreview": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "previewUrl": "https://previewUrl/...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "201fcd21-967a-4f82-8082-5073bd09d31f" == event.event_id
    ts = datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)

    assert ts == event.timestamp
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name

    events: Optional[Dict[str, ImageEventBase]] = event.resource_update_events
    assert events is not None
    assert "sdm.devices.events.CameraClipPreview.ClipPreview" in events
    e: ImageEventBase | None = events[
        "sdm.devices.events.CameraClipPreview.ClipPreview"
    ]
    assert e
    assert isinstance(e, CameraClipPreviewEvent)

    # Serialize the event then unserialize and veirfy everything is still correct
    data = e.as_dict()
    dump = json.dumps(data)
    data = json.loads(dump)

    e = ImageEventBase.parse_event_dict(data)
    assert e
    assert isinstance(e, CameraClipPreviewEvent)

    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id
    assert ts == e.timestamp
    assert (
        "1cd28a21db0705a03f612d1df956444094a8c85a75fd35dfd277a81b5a9ad2a8c18"
        "2a9a7decb2936664d6e1344915850a31fc3c828c3813765d1c73be1e958f3" == e.event_id
    )
    assert "https://previewUrl/..." == e.preview_url
    expire_ts = datetime.datetime(2019, 1, 1, 0, 15, 1, tzinfo=datetime.timezone.utc)
    assert expire_ts == e.expires_at


def test_update_from_events(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraSound.Sound": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                    },
                    "sdm.devices.events.CameraMotion.Motion": {
                        "eventSessionId": "DkX5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "EXXVQVUdGNUlTU2V4MGV2aTNXV...",
                    },
                    "sdm.devices.events.CameraPerson.Person": {
                        "eventSessionId": "EjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FZZVQVUdGNUlTU2V4MGV2aTNXV...",
                    },
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )

    event = event.with_events(
        set({"sdm.devices.events.CameraMotion.Motion", "not-found"})
    )
    assert "0120ecc7-3b57-4eb4-9941-91609f189fb4" == event.event_id
    ts = datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)

    assert ts == event.timestamp
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name
    events = event.resource_update_events
    assert events is not None
    assert len(events) == 1
    assert "sdm.devices.events.CameraMotion.Motion" in events
    e = events["sdm.devices.events.CameraMotion.Motion"]
    assert "EXXVQVUdGNUlTU2V4MGV2aTNXV..." == e.event_id
    assert "DkX5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id
    assert ts == e.timestamp
    expire_ts = datetime.datetime(2019, 1, 1, 0, 0, 31, tzinfo=datetime.timezone.utc)
    assert expire_ts == e.expires_at
    event_token = EventToken.decode(e.event_token)
    assert event_token.event_id == "EXXVQVUdGNUlTU2V4MGV2aTNXV..."
    assert event_token.event_session_id == "DkX5Y3VKaTZwR3o4Y19YbTVfMF..."

    event = event.with_events(set(["unknown"]))
    events = event.resource_update_events
    assert not events


def test_event_image_type() -> None:
    """Test for EventImageType."""
    assert (
        EventImageType.from_string(str(EventImageType.CLIP_PREVIEW))
        == EventImageType.CLIP_PREVIEW
    )
    assert EventImageType.from_string(str(EventImageType.IMAGE)) == EventImageType.IMAGE
    assert EventImageType.from_string("video/mp4") == EventImageType.CLIP_PREVIEW
    assert EventImageType.from_string("image/jpeg") == EventImageType.IMAGE

    assert EventImageType.from_string("image/gif") != EventImageType.IMAGE
    assert EventImageType.from_string("image/gif") == EventImageType.IMAGE_PREVIEW
    assert EventImageType.from_string("image/gif").content_type == "image/gif"
    assert EventImageType.from_string("other").content_type == "other"


def test_event_token() -> None:
    """Test for an EventToken."""
    token = EventToken("session-id", "event-id")
    assert token.event_session_id == "session-id"
    assert token.event_id == "event-id"
    encoded_token = token.encode()
    decoded_token = EventToken.decode(encoded_token)
    assert decoded_token.event_session_id == "session-id"
    assert decoded_token.event_id == "event-id"


def test_decode_token_failure() -> None:
    """Test failure to decode an event token."""
    with pytest.raises(DecodeException):
        EventToken.decode("some-bogus-token")


def test_decode_token_failure_items() -> None:
    """Test failure to decode an event token."""
    data = ["wrong type"]
    b = json.dumps(data).encode("utf-8")
    token = base64.b64encode(b).decode("utf-8")
    with pytest.raises(DecodeException):
        EventToken.decode(token)


def test_decode_token_failure_dict() -> None:
    """Test failure to decode an event token."""
    data = {"a": "b", "c": "d"}
    b = json.dumps(data).encode("utf-8")
    token = base64.b64encode(b).decode("utf-8")
    with pytest.raises(DecodeException):
        EventToken.decode(token)


def test_event_zone(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraPerson.Person": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                        "zones": ["Zone 1"],
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "0120ecc7-3b57-4eb4-9941-91609f189fb4" == event.event_id
    assert (
        datetime.datetime(2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)
        == event.timestamp
    )
    assert "enterprises/project-id/devices/device-id" == event.resource_update_name
    events = event.resource_update_events
    assert events is not None
    assert "sdm.devices.events.CameraPerson.Person" in events
    e = events["sdm.devices.events.CameraPerson.Person"]
    assert "FWWVQVUdGNUlTU2V4MGV2aTNXV..." == e.event_id
    assert "CjY5Y3VKaTZwR3o4Y19YbTVfMF..." == e.event_session_id
    assert e.zones == ["Zone 1"]


def test_unknown_event_type(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    """Test at event published with a type that is not recognized."""
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.Ignored.EventType": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert event.event_id == "0120ecc7-3b57-4eb4-9941-91609f189fb4"
    assert event.timestamp == datetime.datetime(
        2019, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc
    )
    assert event.resource_update_name == "enterprises/project-id/devices/device-id"
    assert event.resource_update_events == {}


def test_event_message_repr(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    """Test at event published with a type that is not recognized."""
    event = fake_event_message(
        {
            "eventId": "0120ecc7-3b57-4eb4-9941-91609f189fb4",
            "timestamp": "2019-01-01T00:00:01Z",
            "resourceUpdate": {
                "name": "enterprises/project-id/devices/device-id",
                "events": {
                    "sdm.devices.events.CameraMotion.Motion": {
                        "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        "eventId": "FWWVQVUdGNUlTU2V4MGV2aTNXV...",
                    }
                },
            },
            "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
        }
    )
    assert "EventMessage{" in repr(event)
    assert event.resource_update_events
    assert "sdm.devices.events.CameraMotion.Motion" in event.resource_update_events
    motion = event.resource_update_events["sdm.devices.events.CameraMotion.Motion"]
    assert motion
    assert "CameraMotionEvent(" in repr(motion)



def test_missing_preview_url(
    fake_event_message: Callable[[Dict[str, Any]], EventMessage]
) -> None:
    with pytest.raises(ValueError, match="EventMessage has invalid value"):
        fake_event_message(
            {
                "eventId": "201fcd21-967a-4f82-8082-5073bd09d31f",
                "timestamp": "2019-01-01T00:00:01Z",
                "resourceUpdate": {
                    "name": "enterprises/project-id/devices/device-id",
                    "events": {
                        "sdm.devices.events.CameraClipPreview.ClipPreview": {
                            "eventSessionId": "CjY5Y3VKaTZwR3o4Y19YbTVfMF...",
                        }
                    },
                },
                "userId": "AVPHwEuBfnPOnTqzVFT4IONX2Qqhu9EJ4ubO-bNnQ-yi",
            }
        )
